"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

from functools import wraps
import uuid
import datetime

from inmanta import data
from inmanta import const
from tornado import gen


VALID_CLIENT_TYPES = ["api", "agent", "compiler", "public"]


def protocol(index=False, id=False, operation="POST", reply=True, arg_options={}, timeout=None, server_agent=False, api=True,
             agent_server=False, validate_sid=None, client_types=["public"]):
    """
        Decorator to identify a method as a RPC call. The arguments of the decorator are used by each transport to build
        and model the protocol.

        :param index: A method that returns a list of resources. The url of this method is only the method/resource name.
        :param id: This method requires an id of a resource. The python function should have an id parameter.
        :param operation: The type of HTTP operation (verb)
        :param timeout: nr of seconds before request it terminated
        :param api This is a call from the client to the Server (True if not server_agent and not agent_server)
        :param server_agent: This is a call from the Server to the Agent (reverse http channel through long poll)
        :param agent_server: This is a call from the Agent to the Server
        :param validate_sid: This call requires a valid session, true by default if agent_server and not api
        :param client_types: The allowed client types for this call
        :param arg_options Options related to arguments passed to the method. The key of this dict is the name of the arg to
            which the options apply. The value is another dict that can contain the following options:
                header: Map this argument to a header with the following name.
                reply_header: If the argument is mapped to a header, this header will also be included in the reply
                getter: Call this method after validation and pass its return value to the method call. This may change the
                        type of the argument. This method can raise an HTTPException to return a 404 for example.
    """
    if api is None:
        api = not server_agent and not agent_server

    if validate_sid is None:
        validate_sid = agent_server and not api

    properties = {
        "index": index,
        "id": id,
        "reply": reply,
        "operation": operation,
        "timeout": timeout,
        "api": api,
        "server_agent": server_agent,
        "agent_server": agent_server,
        "validate_sid": validate_sid,
        "arg_options": arg_options,
        "client_types": client_types,
    }

    def wrapper(func):
        """
            Return the wrapper method so it can be used to do RPC calls. Additionally it adds the protocol properties of the
            decorator to the method.
        """
        @wraps(func)
        @gen.coroutine
        def wrapped_method(self, *args, **kwargs):
            """
                This wrapper will first call the original method to validate the arguments passed and possible custom
                validation logic in the method itself.
            """
            func(self, *args, **kwargs)

            result = yield self._call(args=args, kwargs=kwargs, protocol_properties=properties)
            return result

        properties["method"] = func
        wrapped_method.__protocol_properties__ = properties
        return wrapped_method

    return wrapper


class HTTPException(Exception):
    def __init__(self, code, message=None):
        super().__init__(code, message)
        self.code = code
        self.message = message


class Method(object):
    """
        A decorator to add methods to the protocol.
    """
    __method_name__ = None


@gen.coroutine
def get_environment(env: uuid.UUID, metadata: dict) -> data.Environment:
    metadata[const.INMANTA_URN + "env"] = str(env)
    env = yield data.Environment.get_by_id(env)
    if env is None:
        raise HTTPException(code=404, message="The given environment id does not exist!")
    return env


@gen.coroutine
def add_env(env: uuid.UUID, metadata: dict) -> uuid.UUID:
    metadata[const.INMANTA_URN + "env"] = str(env)
    return env


@gen.coroutine
def ignore_env(obj: object, metadata: dict) -> object:
    """
        This mapper only adds an env all for authz
    """
    metadata[const.INMANTA_URN + "env"] = "all"
    return obj


ENV_ARG = {"header": "X-Inmanta-tid", "getter": get_environment, "reply_header": True}
ENV_OPTS = {"tid": ENV_ARG}
AGENT_ENV_ARG = {"header": "X-Inmanta-tid", "reply_header": True, "getter": add_env}
AGENT_ENV_OPTS = {"tid": AGENT_ENV_ARG}


class Project(Method):
    """
        Method for working with projects
    """
    __method_name__ = "project"

    @protocol(operation="PUT", client_types=["api"])
    def create_project(self, name):
        """
            Create a new project
        """

    @protocol(operation="POST", id=True, client_types=["api"])
    def modify_project(self, id: uuid.UUID, name: str):
        """
            Modify the given project
        """

    @protocol(operation="DELETE", id=True, client_types=["api"])
    def delete_project(self, id: uuid.UUID):
        """
            Delete the given project and all related data
        """

    @protocol(operation="GET", index=True, client_types=["api"])
    def list_projects(self):
        """
            Create a list of projects
        """

    @protocol(operation="GET", id=True, client_types=["api"])
    def get_project(self, id: uuid.UUID):
        """
            Get a project and a list of the ids of all environments
        """


class Environment(Method):
    """
        Method for working with environments
    """
    __method_name__ = "environment"

    @protocol(operation="PUT", client_types=["api"])
    def create_environment(self, project_id: uuid.UUID, name: str, repository: str=None, branch: str=None):
        """
            Create a new environment

            :param project_id The id of the project this environment belongs to
            :param name The name of the environment
            :param repository The url (in git form) of the repository
            :param branch The name of the branch in the repository
        """

    @protocol(operation="POST", id=True, client_types=["api"])
    def modify_environment(self, id: uuid.UUID, name: str, repository: str=None, branch: str=None):
        """
            Modify the given environment

            :param id The id of the environment
            :param name The name of the environment
            :param repository The url (in git form) of the repository
            :param branch The name of the branch in the repository
        """

    @protocol(operation="DELETE", id=True, client_types=["api"])
    def delete_environment(self, id: uuid.UUID):
        """
            Delete the given environment and all related data
        """

    @protocol(operation="GET", index=True, client_types=["api"])
    def list_environments(self):
        """
            Create a list of environments
        """

    @protocol(operation="GET", id=True, client_types=["api"], arg_options={"id": {"getter": add_env}})
    def get_environment(self, id: uuid.UUID, versions: int=None, resources: int=None):
        """
            Get an environment and all versions associated

            :param id The id of the environment to return
            :param versions Include this many available version for this environment.
            :param resources Include this many available resources for this environment.
        """


class EnvironmentSettings(Method):
    """
        Method for listing/getting/setting/removing settings of an environment. This API is also used by agents to configure
        environments.
    """
    __method_name__ = "environment_settings"

    @protocol(operation="GET", index=True, arg_options=ENV_OPTS, api=True, agent_server=True,
              client_types=["api", "agent", "compiler"])
    def list_settings(self, tid: uuid.UUID):
        """
            List the settings in the current environment
        """

    @protocol(operation="POST", id=True, arg_options=ENV_OPTS, api=True, agent_server=True,
              client_types=["api", "agent", "compiler"])
    def set_setting(self, tid: uuid.UUID, id: str, value: object):
        """
            Set a value
        """

    @protocol(operation="GET", id=True, arg_options=ENV_OPTS, api=True, agent_server=True,
              client_types=["api", "agent"])
    def get_setting(self, tid: uuid.UUID, id: str):
        """
            Get a value
        """

    @protocol(operation="DELETE", id=True, arg_options=ENV_OPTS, api=True, agent_server=True,
              client_types=["api", "agent"])
    def delete_setting(self, tid: uuid.UUID, id: str):
        """
            Delete a value
        """


class EnvironmentAuth(Method):
    """
        Method for listing and creating auth tokens for an environment that can be used by the agent and compilers
    """
    __method_name__ = "environment_auth"

    @protocol(operation="POST", index=True, arg_options=ENV_OPTS, client_types=["api", "compiler"])
    def create_token(self, tid: uuid.UUID, client_types: list, idempotent: bool=True):
        """
            Create or get a new token for the given client types. Tokens generated with this call are scoped to the current
            environment.

            :param tid: The environment id
            :param client_types: The client types for which this token is valid (api, agent, compiler)
            :param idempotent: The token should be idempotent, such tokens do not have an expire or issued at set so their
                               value will not change.
        """


class Decommision(Method):
    """
        Decomission an environment
    """
    __method_name__ = "decommission"

    @protocol(operation="POST", id=True, arg_options={"id": {"getter": get_environment}}, client_types=["api"])
    def decomission_environment(self, id: uuid.UUID, metadata: dict=None):
        """
            Decommision an environment. This is done by uploading an empty model to the server and let purge_on_delete handle
            removal.
        """

    @protocol(operation="DELETE", id=True, arg_options={"id": {"getter": get_environment}}, client_types=["api"])
    def clear_environment(self, id: uuid.UUID):
        """
            Clear all data from this environment
        """


class HeartBeatMethod(Method):
    """
        Send a heartbeat to indicate that an agent is alive
    """
    __method_name__ = "heartbeat"

    @protocol(operation="POST", agent_server=True, validate_sid=False, arg_options=ENV_OPTS, client_types=["agent"])
    def heartbeat(self, sid: uuid.UUID, tid: uuid.UUID, endpoint_names: list, nodename: str):
        """
            Send a heartbeat to the server

            :paran sid: The session ID used by this agent at this moment
            :param tid: The environment this node and its agents belongs to
            :param endpoint_names: The names of the endpoints on this node
            :param nodename: The name of the node from which the heart beat comes

            also registered as API method, because it is called with an invalid SID the first time
        """

    @protocol(operation="PUT", agent_server=True, arg_options={"sid": {"getter": ignore_env}}, client_types=["agent"])
    def heartbeat_reply(self, sid: uuid.UUID, reply_id: uuid.UUID, data: dict):
        """
            Send a reply back to the server

            :param sid: The session ID used by this agent at this moment
            :param reply_id: The id data is a reply to
            :param data: The data as a response to the reply
        """


class FileMethod(Method):
    """
        Upload, retrieve and check for file. A file is identified by a hash of its content.
    """
    __method_name__ = "file"

    @protocol(operation="PUT", id=True, agent_server=True, api=True, client_types=["api", "agent", "compiler"],
              arg_options={"id": {"getter": ignore_env}})
    def upload_file(self, id: str, content: str):
        """
            Upload a new file

            :param id The id of the file
            :param content The base64 encoded content of the file
        """

    @protocol(operation="HEAD", id=True, agent_server=True, api=True, client_types=["api", "agent", "compiler"],
              arg_options={"id": {"getter": ignore_env}})
    def stat_file(self, id: str):
        """
            Does the file exist

            :param id The id of the file to check
        """

    @protocol(operation="GET", id=True, agent_server=True, api=True, client_types=["api", "agent", "compiler"],
              arg_options={"id": {"getter": ignore_env}})
    def get_file(self, id: str):
        """
            Retrieve a file

            :param id: The id of the file to retrieve
        """

    @protocol(api=True, client_types=["api", "agent", "compiler"], arg_options={"files": {"getter": ignore_env}})
    def stat_files(self, files: list):
        """
            Check which files exist in the given list

            :param files: A list of file id to check
            :return: A list of files that do not exist.
        """


class ResourceMethod(Method):
    """
        Manage resources on the server
    """
    __method_name__ = "resource"

    @protocol(operation="GET", id=True, agent_server=True, validate_sid=False, arg_options=ENV_OPTS, api=True,
              client_types=["api", "agent"])
    def get_resource(self, tid: uuid.UUID, id: str, logs: bool=None, status: bool=None,
                     log_action: const.ResourceAction=None, log_limit: int=0):
        """
            Return a resource with the given id.

            :param tid The id of the environment this resource belongs to
            :param id Get the resource with the given id
            :param logs Include the logs in the response
            :param status return only resources of this status
            :param log_action The log action to include, leave empty/none for all actions. Valid actions are one of
                              the action strings in const.ResourceAction
            :param log_limit Limit the number of logs included in the response
        """

    @protocol(operation="GET", index=True, agent_server=True, arg_options=ENV_OPTS, client_types=["agent"])
    def get_resources_for_agent(self, tid: uuid.UUID, agent: str, version: int=None):
        """
            Return the most recent state for the resources associated with agent, or the version requested

            :param tid: The id of the environment this resource belongs to
            :param agent: The agent
            :param version: The version to retrieve. If none, the latest available version is returned. With a specific version
                            that version is returned, even if it has not been released yet.
        """

    @protocol(operation="POST", index=True, agent_server=True, arg_options=ENV_OPTS, client_types=["agent"])
    def resource_action_update(self, tid: uuid.UUID, resource_ids: list, action_id: uuid.UUID, action: const.ResourceAction,
                               started: datetime.datetime=None, finished: datetime.datetime=None,
                               status: const.ResourceState=None, messages: list=[], changes: dict={},
                               change: const.Change=None, send_events: bool=False):
        """
            Send a resource update to the server

            :param tid: The id of the environment this resource belongs to
            :param resource_ids: The resource with the given resource_version_id id from the agent
            :param action_id: A unique id to indicate the resource action that has be updated
            :param action: The action performed
            :param started: The timestamp when this action was started. When this action (action_id) has not been saved yet,
                            started has to be defined.
            :param finished: The timestamp when this action was finished. Afterwards, no changes with the same action_id
                            can be stored. The status field also has to be set.
            :param status: The current status of the resource (if known)
            :param messages: A list of log entries to add to this entry.
            :param change:s A dict of changes to this resource. The key of this dict indicates the attributes/fields that
                           have been changed. The value contains the new value and/or the original value.
            :param change: The result of the changes
            :param send_events: Send events to the dependents of this resource
        """


class VersionMethod(Method):
    """
        Manage configuration model versions
    """
    __method_name__ = "version"

    @protocol(index=True, operation="GET", arg_options=ENV_OPTS, client_types=["api"])
    def list_versions(self, tid: uuid.UUID, start: int=None, limit: int=None):
        """
            Returns a list of all available versions

            :param tid: The id of the environment
            :param start: Optional, parameter to control the amount of results that are returned. 0 is the latest version.
            :param limit: Optional, parameter to control the amount of results returned.
        """

    @protocol(operation="GET", id=True, arg_options=ENV_OPTS, client_types=["api"])
    def get_version(self, tid: uuid.UUID, id: int, include_logs: bool=None, log_filter: str=None, limit: int=None):
        """
            Get a particular version and a list of all resources in this version

            :param tid: The id of the environment
            :param id: The id of the version to retrieve
            :param include_logs: If true, a log of all operations on all resources is included
            :param log_filter: Filter log to only include actions of the specified type
            :param limit: The maximal number of actions to return per resource (starting from the latest)
        """

    @protocol(operation="DELETE", id=True, arg_options=ENV_OPTS, client_types=["api"])
    def delete_version(self, tid: uuid.UUID, id: int):
        """
            Delete a particular version and resources

            :param tid: The id of the environment
            :param id: The id of the version to retrieve
        """

    @protocol(operation="PUT", arg_options=ENV_OPTS, client_types=["compiler"])
    def put_version(self, tid: uuid.UUID, version: int, resources: list, resource_state: dict={}, unknowns: list=None,
                    version_info: dict=None):
        """
            Store a new version of the configuration model

            :param tid: The id of the environment
            :param version: The version of the configuration model
            :param resources: A list of all resources in the configuration model (deployable)
            :param resource_state: A dictionary with the initial const.ResourceState per resource id
            :param unknowns: A list of unknown parameters that caused the model to be incomplete
            :param version_info: Module version information
        """

    @protocol(operation="POST", id=True, arg_options=ENV_OPTS, client_types=["api", "compiler"])
    def release_version(self, tid: uuid.UUID, id: int, push: bool):
        """
            Release version of the configuration model for deployment.

            :param tid: The id of the environment
            :param id: The version of the CM to deploy
            :param push: Notify all agents to deploy the version
        """


class DryRunMethod(Method):
    """
        Method for requesting and quering a dryrun
    """
    __method_name__ = "dryrun"

    @protocol(operation="POST", id=True, arg_options=ENV_OPTS, client_types=["api"])
    def dryrun_request(self, tid: uuid.UUID, id: int):
        """
            Do a dryrun

            :param tid: The id of the environment
            :param id: The version of the CM to deploy
        """

    @protocol(operation="GET", arg_options=ENV_OPTS, client_types=["api"])
    def dryrun_list(self, tid: uuid.UUID, version: int=None):
        """
            Create a list of dry runs

            :param tid: The id of the environment
            :param version: Only for this version
        """

    @protocol(operation="GET", id=True, arg_options=ENV_OPTS, client_types=["api"])
    def dryrun_report(self, tid: uuid.UUID, id: uuid.UUID):
        """
            Create a dryrun report

            :param tid: The id of the environment
            :param id: The version dryrun to report
        """

    @protocol(operation="PUT", id=True, agent_server=True, arg_options=ENV_OPTS, client_types=["agent"])
    def dryrun_update(self, tid: uuid.UUID, id: uuid.UUID, resource: str, changes: dict):
        """
            Store dryrun results at the server

            :param tid: The id of the environment
            :param id: The version dryrun to report
            :param resource: The id of the resource
            :param changes: The required changes
        """


class AgentDryRun(Method):
    """
        Method for requesting a dryrun from an agent
    """
    __method_name__ = "agent_dryrun"

    @protocol(operation="POST", id=True, server_agent=True, timeout=5, arg_options=AGENT_ENV_OPTS, client_types=[])
    def do_dryrun(self, tid: uuid.UUID, id: uuid.UUID, agent: str, version: int):
        """
            Do a dryrun on an agent

            :param tid: The environment id
            :param id: The id of the dryrun
            :param agent: The agent to do the dryrun for
            :param version: The version of the model to dryrun
        """


class NotifyMethod(Method):
    """
        Method to notify the server of changes in the configuration model source code
    """
    __method_name__ = "notify"

    @protocol(operation="GET", id=True, arg_options={"id": {"getter": get_environment}}, client_types=["api"])
    def notify_change_get(self, id: uuid.UUID, update: bool=True):
        """
            Simplified GET version of the POST method
        """

    @protocol(operation="POST", id=True, arg_options={"id": {"getter": get_environment}}, client_types=["api"])
    def notify_change(self, id: uuid.UUID, update: bool=True, metadata: dict={}):
        """
            Notify the server that the repository of the environment with the given id, has changed.

            :param id: The id of the environment
            :param update: Update the model code and modules. Default value is true
            :param metadata: The metadata that indicates the source of the compilation trigger.
        """

    @protocol(operation="HEAD", id=True, client_types=["api"])
    def is_compiling(self, id: uuid.UUID):
        """
           Is a compiler running for the given environment

           :param id: The environment id
        """


class ParameterMethod(Method):
    """
        Get and set parameters on the server
    """
    __method_name__ = "parameter"

    @protocol(operation="GET", id=True, arg_options=ENV_OPTS, client_types=["api", "compiler", "agent"])
    def get_param(self, tid: uuid.UUID, id: str, resource_id: str=None):
        """
            Get a parameter from the server.

            :param tid: The id of the environment
            :param id: The name of the parameter
            :param resource_id: Optionally, scope the parameter to resource (fact),
                                if the resource id should not contain a version, the latest version is used
            :return: Returns the following status codes:
                    200: The parameter content is returned
                    404: The parameter is not found and unable to find it because its resource is not known to the server
                    410: The parameter has expired
                    503: The parameter is not found but its value is requested from an agent
        """

    @protocol(operation="PUT", id=True, arg_options=ENV_OPTS, client_types=["api", "compiler", "agent"])
    def set_param(self, tid: uuid.UUID, id: str, source: str, value: str, resource_id: str=None, metadata: dict={},
                  recompile: bool=False):
        """
            Set a parameter on the server. If the parameter is an tracked unknown, it will trigger a recompile on the server.
            Otherwise, if the value is changed and recompile is true, a recompile is also triggered.

            :param tid: The id of the environment
            :param id: The name of the parameter
            :param resource_id: Optionally, scope the parameter to resource (fact)
            :param source: The source of the parameter, this can be the user, agent, plugin, compiler, ...
            :param value: The value of the parameter
            :param metadata: metadata about the parameter
            :param recompile: Whether to trigger a recompile
        """

    @protocol(operation="DELETE", id=True, arg_options=ENV_OPTS, client_types=["api", "compiler", "agent"])
    def delete_param(self, tid: uuid.UUID, id: str, resource_id: str=None):
        """
            Delete a parameter on the server

            :param tid: The id of the environment
            :param id: The name of the parameter
            :param resource_id: The resource id of the parameter
        """

    @protocol(operation="POST", index=True, arg_options=ENV_OPTS, client_types=["api", "compiler"])
    def list_params(self, tid: uuid.UUID, query: dict={}):
        """
            List/query parameters in this environment

            :param tid: The id of the environment
            :param query: A query to match against metadata
        """


class ParametersMethod(Method):
    """
        Get and set parameters on the server
    """
    __method_name__ = "parameters"

    @protocol(operation="PUT", index=True, agent_server=True, arg_options=ENV_OPTS, client_types=["api", "compiler", "agent"])
    def set_parameters(self, tid: uuid.UUID, parameters: list):
        """
            Set a parameter on the server

            :param tid: The id of the environment
            :param parameters: A list of dicts with the following keys:
                - id The name of the parameter
                - source The source of the parameter, this can be the user, agent, plugin, compiler, ...
                - value The value of the parameter
                - resource_id Optionally, scope the parameter to resource (fact)
                - metadata metadata about the parameter
        """


class AgentParameterMethod(Method):
    """
        Get parameters from the agent
    """
    __method_name__ = "agent_parameter"

    @protocol(operation="POST", server_agent=True, timeout=5, arg_options=AGENT_ENV_OPTS, client_types=[])
    def get_parameter(self, tid: uuid.UUID, agent: str, resource: dict):
        """
            Get all parameters/facts known by the agents for the given resource

            :param tid: The environment
            :param agent: The agent get the parameters froms
            :param resource: The resource to query the parameters from
        """


class FormMethod(Method):
    """
        Methods for creating and manipulating forms
    """
    __method_name__ = "form"

    @protocol(operation="GET", index=True, arg_options=ENV_OPTS, client_types=["api", "compiler"])
    def list_forms(self, tid: uuid.UUID):
        """
            List all available forms in an environment
        """

    @protocol(operation="GET", id=True, arg_options=ENV_OPTS, client_types=["api", "compiler"])
    def get_form(self, tid: uuid.UUID, id: str):
        """
            Get a form
        """

    @protocol(operation="PUT", id=True, arg_options=ENV_OPTS, client_types=["api", "compiler"])
    def put_form(self, tid: uuid.UUID, id: str, form: dict):
        """
            Upload a form
        """


class FormRecords(Method):
    """
        Methods for working with form records
    """
    __method_name__ = "records"

    @protocol(operation="GET", index=True, arg_options=ENV_OPTS, client_types=["api", "compiler"])
    def list_records(self, tid: uuid.UUID, form_type: str, include_record: bool=False):
        """
            Get a list of all records of a specific form

            :param tid: The id of the environment
            :param form_type: The type of the form
            :param include_record: Include all the data contained in the record as well
        """

    @protocol(operation="GET", id=True, arg_options=ENV_OPTS, client_types=["api", "compiler"])
    def get_record(self, tid: uuid.UUID, id: uuid.UUID):
        """
            Get a record from the server

            :param tid: The id of the environment
            :param record_id: The id of the record
        """

    @protocol(operation="PUT", id=True, arg_options=ENV_OPTS, client_types=["api", "compiler"])
    def update_record(self, tid: uuid.UUID, id: uuid.UUID, form: dict):
        """
            Update a record

            :param tid: The id of the environment
            :param id: The id of the record
            :param form: The values of each field
        """

    @protocol(operation="POST", index=True, arg_options=ENV_OPTS, client_types=["api", "compiler"])
    def create_record(self, tid: uuid.UUID, form_type: str, form: dict):
        """
            Get a list of all records of a specific form

            :param tid: The id of the environment
            :param form_type: The type of the form
            :param form: The values for each field
        """

    @protocol(operation="DELETE", id=True, arg_options=ENV_OPTS, client_types=["api", "compiler"])
    def delete_record(self, tid: uuid.UUID, id: uuid.UUID):
        """
            Delete a record

            :param tid: The id of the environment
            :param id: The id of the record
        """


class CodeMethod(Method):
    """
        Upload code to the server
    """
    __method_name__ = "code"

    @protocol(operation="PUT", id=True, arg_options=ENV_OPTS, client_types=["compiler"])
    def upload_code(self, tid: uuid.UUID, id: int, resource: str, sources: dict):
        """
            Upload the supporting code to the server

            :param tid: The environment the code belongs to
            :param id: The id (version) of the configuration model
            :param sources: The source files that contain handlers and inmanta plug-ins
                {code_hash:(file_name, provider.__module__, source_code, [req])}
        """

    @protocol(operation="GET", id=True, agent_server=True, arg_options=ENV_OPTS, client_types=["agent"])
    def get_code(self, tid: uuid.UUID, id: int, resource: str):
        """
            Get the code for a given version of the configuration model

            :param tid: The environment the code belongs to
            :param id: The id (version) of the configuration model
        """


class CodeBatchedMethod(Method):
    """
        Upload code to the server
    """
    __method_name__ = "codebatched"

    @protocol(operation="PUT", id=True, arg_options=ENV_OPTS, client_types=["compiler"])
    def upload_code_batched(self, tid: uuid.UUID, id: int, resources: dict):
        """
            Upload the supporting code to the server

            :param tid: The environment the code belongs to
            :param id: The id (version) of the configuration model
            :param resource: a dict mapping resources to dicts mapping file names to file hashes
        """


class FileDiff(Method):
    """
        Generate download the diff of two hashes
    """
    __method_name__ = "filediff"

    @protocol(client_types=["api"])
    def diff(self, a: str, b: str):
        """
            Returns the diff of the files with the two given ids
        """


class CompileReport(Method):
    """
        Get a list of compile reports
    """
    __method_name__ = "compilereport"

    @protocol(operation="GET", index=True, arg_options=ENV_OPTS, client_types=["api"])
    def get_reports(self, tid: uuid.UUID, start: str=None, end: str=None, limit: int=None):
        """
            Return compile reports newer then start

            :param tid: The id of the environment to get a report from
            :param start: Reports after start
            :param end: Reports before end
            :param limit: Maximum number of results
        """

    @protocol(operation="GET", id=True, client_types=["api"])
    def get_report(self, id: uuid.UUID):
        """
            Get a compile report from the server

            :param compile_id: The id of the compile and its reports to fetch.
        """


class Snapshot(Method):
    """
        Methods to create and restore snapshots
    """
    __method_name__ = "snapshot"

    @protocol(operation="GET", index=True, arg_options=ENV_OPTS, client_types=["api"])
    def list_snapshots(self, tid: uuid.UUID):
        """
            Create a list of all snapshots
        """

    @protocol(operation="GET", id=True, arg_options=ENV_OPTS, client_types=["api"])
    def get_snapshot(self, tid: uuid.UUID, id: uuid.UUID):
        """
            Get details about a snapshot and a list of resources that can be snapshot-ed
        """

    @protocol(operation="POST", index=True, arg_options=ENV_OPTS, client_types=["api"])
    def create_snapshot(self, tid: uuid.UUID, name: str=None):
        """
            Request a new snapshot
        """

    @protocol(operation="PUT", id=True, agent_server=True, arg_options=ENV_OPTS, client_types=["agent"])
    def update_snapshot(self, tid: uuid.UUID, id: uuid.UUID, resource_id: str, snapshot_data: str, error: bool, success: bool,
                        start: datetime.datetime, stop: datetime.datetime, size: int, msg: str=None):
        """
            Update a snapshot with the details of a specific resource

            :param tid :The environment the snapshot is created in
            :param id: The id of the snapshot to update
            :param resource_id: The id of the resource of which a snapshot was created
            :param snapshot_data: The hash of the snapshot data that was uploaded to the fileserver
            :param start: The time when the snapshot was started
            :param stop: The time the snapshot ended
            :param size: The size of the snapshot in bytes
            :param msg: An optional message about the snapshot
        """

    @protocol(operation="DELETE", id=True, arg_options=ENV_OPTS, client_types=["api"])
    def delete_snapshot(self, tid: uuid.UUID, id: uuid.UUID):
        """
            Delete a snapshot
        """


class AgentSnapshot(Method):
    """
        Snapshot operations performed on the agent
    """
    __method_name__ = "agent_snapshot"

    @protocol(operation="POST", server_agent=True, timeout=5, arg_options=AGENT_ENV_OPTS, client_types=[])
    def do_snapshot(self, tid: uuid.UUID, agent: str, snapshot_id: uuid.UUID, resources: list):
        """
            Create a snapshot of the requested resource

            :param tid The environment the snapshot is created in
            :param agent The name of the agent
            :param snapshot_id The id of the snapshot to create
            :param resource A list of resources to snapshot
        """


class RestoreSnapshot(Method):
    """
        Restore a snapshot
    """
    __method_name__ = "restore"

    @protocol(operation="POST", index=True, arg_options=ENV_OPTS, client_types=["api"])
    def restore_snapshot(self, tid: uuid.UUID, snapshot: uuid.UUID):
        """
            Restore a snapshot

            :param tid: The environment the snapshot is created in
            :param snapshot: The id of the snapshot to restore
        """

    @protocol(operation="POST", id=True, agent_server=True, arg_options=ENV_OPTS, client_types=["agent"])
    def update_restore(self, tid: uuid.UUID, id: uuid.UUID, resource_id: str, success: bool, error: bool, msg: str,
                       start: datetime.datetime, stop: datetime.datetime):
        """
            Update the status of a restore

            :param tid: The environment to restore the snapshot in
            :param id: The id of the restore
            :param resource_id: The state id of the resource that was restored.
        """

    @protocol(operation="GET", index=True, arg_options=ENV_OPTS, client_types=["api"])
    def list_restores(self, tid: uuid.UUID):
        """
            List finished and ongoing restores
        """

    @protocol(operation="GET", id=True, arg_options=ENV_OPTS, client_types=["api"])
    def get_restore_status(self, tid: uuid.UUID, id: uuid.UUID):
        """
            Get the status of a restore
        """

    @protocol(operation="DELETE", id=True, arg_options=ENV_OPTS, client_types=["api"])
    def delete_restore(self, tid: uuid.UUID, id: uuid.UUID):
        """
            Cancel a restore
        """


class AgentRestore(Method):
    """
        Snapshot operations performed on the agent
    """
    __method_name__ = "agent_restore"

    @protocol(operation="POST", server_agent=True, timeout=5, arg_options=AGENT_ENV_OPTS, client_types=[])
    def do_restore(self, tid: uuid.UUID, agent: str, restore_id: uuid.UUID, snapshot_id: uuid.UUID, resources: list):
        """
            Create a snapshot of the requested resource

            :param tid: The environment the snapshot is created in
            :param agent: The name of the agent
            :param restore_id: THe id of the restore operation
            :param snapshot_id: The id of the snapshot to restore
            :param resource: A list of resources to restore
        """


class NodeMethod(Method):
    """
        Get a list of all agents
    """
    __method_name__ = "agentproc"

    @protocol(operation="GET", index=True, client_types=["api"])
    def list_agent_processes(self, environment: uuid.UUID=None, expired: bool=True):
        """
            Return a list of all nodes and the agents for these nodes

            :param environment: An optional environment. If set, only the agents that belong to this environment are returned
            :param all: Optional, also show expired.
            :return: A list of nodes
        """

    @protocol(operation="GET", id=True, client_types=["api"])
    def get_agent_process(self, id: uuid.UUID):
        """
            Return a detailed report for a node

            :param agentid: The id of the node
            :return: The requested node
        """


class ServerAgentApiMethod(Method):
    """
        Get a list of all agents
    """
    __method_name__ = "agent"

    @protocol(operation="POST", id=True, api=True, timeout=5, arg_options=ENV_OPTS, client_types=["api"])
    def trigger_agent(self, tid: uuid.UUID, id: str):
        """
            Request the server to reload an agent

            :param tid The environment this agent is defined in
            :param id The name of the agent
            :return The requested node
        """

    @protocol(operation="GET", api=True, timeout=5, arg_options=ENV_OPTS, client_types=["api"])
    def list_agents(self, tid: uuid.UUID):
        """
            List all agent for an environment

            :param tid The environment the agents are defined in
        """


class AgentReporting(Method):
    """
        Reporting by the agent to the server
    """
    __method_name__ = "status"

    @protocol(operation="GET", server_agent=True, timeout=5, client_types=[])
    def get_status(self):
        """
            A call from the server to the agent to report its status to the server

            :return: A map with report items
        """


class AgentState(Method):
    """
        Methods to allow the server to set the agents state
    """
    __method_name__ = "agentstate"

    @protocol(operation="POST", server_agent=True, timeout=5, client_types=[])
    def set_state(self, agent: str, enabled: bool):
        """
            Set the state of the agent.
        """

    @protocol(operation="POST", id=True, server_agent=True, timeout=5, arg_options=AGENT_ENV_OPTS, client_types=[])
    def trigger(self, tid: uuid.UUID, id: str):
        """
            Request an agent to reload resources

            :param tid The environment this agent is defined in
            :param id The name of the agent
        """


class AgentResourceEvent(Method):
    """
        Methods to send event to the server
    """
    __method_name__ = "event"

    @protocol(operation="PUT", id=True, server_agent=True, timeout=5, arg_options=AGENT_ENV_OPTS, client_types=[])
    def resource_event(self, tid: uuid.UUID, id: str, resource: str, send_events: bool,
                       state: const.ResourceState, change: const.Change, changes={}):
        """
            Tell an agent a resource it waits for has been updated

            :param tid The environment this agent is defined in
            :param id The name of the agent
            :param resource The resource ID of the resource being updated
            :param send_events Does the resource have send_events enabled?
            :param state State the resource acquired (deployed, skipped, canceled)
            :param change The change that was made to the resource
            :param changes The changes made to the resource
        """


class AgentRecovery(Method):
    """
        Methods for the agent to get its initial state from the server
    """
    __method_name__ = "agentrecovery"

    @protocol(operation="GET", agent_server=True, arg_options=ENV_OPTS, client_types=["agent"])
    def get_state(self, tid: uuid.UUID, sid: uuid.UUID, agent: str):
        """
            Get the state for this agent.

            returns a map
            {
             enabled: bool
            }
        """



"""
    Copyright 2018 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
from tornado.ioloop import IOLoop
from inmanta.server import server
from inmanta.server.protocol import RESTServer
from inmanta.server.agentmanager import AgentManager


class InmantaBootloader(object):

    def __init__(self, agent_no_log=False):
        self.restserver = RESTServer()
        self.agent_no_log = agent_no_log

    def get_server_slice(self):
        io_loop = IOLoop.current()
        return server.Server(io_loop, agent_no_log=self.agent_no_log)

    def get_agent_manager_slice(self):
        return AgentManager(self.restserver)

    def get_server_slices(self):
        return [self.get_server_slice(), self.get_agent_manager_slice()]

    def start(self):
        for mypart in self.get_server_slices():
            self.restserver.add_endpoint(mypart)
        self.restserver.start()

    def stop(self):
        self.restserver.stop()



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
from inmanta import protocol
import time


def test_jwt_create(inmanta_config):
    """
        Test creating, signing and verifying JWT with HS256 from the configuration
    """
    jot = protocol.encode_token(["api"])
    payload = protocol.decode_token(jot)

    assert "api" in payload["urn:inmanta:ct"]

    # test creating an idempotent token
    jot1 = protocol.encode_token(["agent"], idempotent=True)
    jot3 = protocol.encode_token(["agent"])
    time.sleep(1)
    jot2 = protocol.encode_token(["agent"], idempotent=True)
    jot4 = protocol.encode_token(["agent"])
    assert jot1 == jot2
    assert jot3 != jot4
    assert jot1 != jot3
    assert jot2 != jot3



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
from collections import defaultdict, namedtuple
import time
import json
import uuid
from threading import Condition
from itertools import groupby
import logging
import os
import shutil
import subprocess


from tornado import gen
import pytest
from _pytest.fixtures import fixture

from inmanta import agent, data, const, execute, config
from inmanta.agent.handler import provider, ResourceHandler, SkipResource, HandlerContext
from inmanta.resources import resource, Resource
import inmanta.agent.agent
from inmanta.agent.agent import Agent
from utils import retry_limited, assert_equal_ish, UNKWN
from inmanta.config import Config
from inmanta.ast import CompilerException
from inmanta.server.bootloader import InmantaBootloader
from inmanta.server import SLICE_AGENT_MANAGER
from typing import List, Tuple, Optional, Dict
from inmanta.const import ResourceState

logger = logging.getLogger("inmanta.test.server_agent")

ResourceContainer = namedtuple('ResourceContainer', ['Provider', 'waiter', 'wait_for_done_with_waiters'])


@fixture(scope="function")
def resource_container():

    @resource("test::Resource", agent="agent", id_attribute="key")
    class MyResource(Resource):
        """
            A file on a filesystem
        """
        fields = ("key", "value", "purged", "state_id", "allow_snapshot", "allow_restore")

    @resource("test::Fact", agent="agent", id_attribute="key")
    class FactResource(Resource):
        """
            A file on a filesystem
        """
        fields = ("key", "value", "purged", "skip", "factvalue", 'skipFact')

    @resource("test::Fail", agent="agent", id_attribute="key")
    class FailR(Resource):
        """
            A file on a filesystem
        """
        fields = ("key", "value", "purged", "state_id", "allow_snapshot", "allow_restore")

    @resource("test::Wait", agent="agent", id_attribute="key")
    class WaitR(Resource):
        """
            A file on a filesystem
        """
        fields = ("key", "value", "purged", "state_id", "allow_snapshot", "allow_restore")

    @resource("test::Noprov", agent="agent", id_attribute="key")
    class NoProv(Resource):
        """
            A file on a filesystem
        """
        fields = ("key", "value", "purged", "state_id", "allow_snapshot", "allow_restore")

    @resource("test::FailFast", agent="agent", id_attribute="key")
    class FailFastR(Resource):
        """
            A file on a filesystem
        """
        fields = ("key", "value", "purged", "state_id", "allow_snapshot", "allow_restore")

    @resource("test::BadEvents", agent="agent", id_attribute="key")
    class BadeEventR(Resource):
        """
            A file on a filesystem
        """
        fields = ("key", "value", "purged", "state_id", "allow_snapshot", "allow_restore")

    @provider("test::Resource", name="test_resource")
    class Provider(ResourceHandler):

        def check_resource(self, ctx, resource):
            self.read(resource.id.get_agent_name(), resource.key)
            assert resource.value != const.UNKNOWN_STRING
            current = resource.clone()
            current.purged = not self.isset(resource.id.get_agent_name(), resource.key)

            if not current.purged:
                current.value = self.get(resource.id.get_agent_name(), resource.key)
            else:
                current.value = None

            return current

        def do_changes(self, ctx, resource, changes):
            if self.skip(resource.id.get_agent_name(), resource.key):
                raise SkipResource()

            if self.fail(resource.id.get_agent_name(), resource.key):
                raise Exception("Failed")

            self.touch(resource.id.get_agent_name(), resource.key)

            if "purged" in changes:
                if changes["purged"]["desired"]:
                    self.delete(resource.id.get_agent_name(), resource.key)
                    ctx.set_purged()
                else:
                    self.set(resource.id.get_agent_name(), resource.key, resource.value)
                    ctx.set_created()

            elif "value" in changes:
                self.set(resource.id.get_agent_name(), resource.key, resource.value)
                ctx.set_updated()

            return changes

        def snapshot(self, resource):
            return json.dumps({"value": self.get(resource.id.get_agent_name(), resource.key), "metadata": "1234"}).encode()

        def restore(self, resource, snapshot_id):
            content = self.get_file(snapshot_id)
            if content is None:
                return

            data = json.loads(content.decode())
            if "value" in data:
                self.set(resource.id.get_agent_name(), resource.key, data["value"])

        def facts(self, ctx, resource):
            return {"length": len(self.get(resource.id.get_agent_name(), resource.key)), "key1": "value1", "key2": "value2"}

        def can_process_events(self) -> bool:
            return True

        def process_events(self, ctx, resource, events):
            self.__class__._EVENTS[resource.id.get_agent_name()][resource.key].append(events)
            super(Provider, self).process_events(ctx, resource, events)

        def can_reload(self) -> bool:
            return True

        def do_reload(self, ctx, resource):
            self.__class__._RELOAD_COUNT[resource.id.get_agent_name()][resource.key] += 1

        _STATE = defaultdict(dict)
        _WRITE_COUNT = defaultdict(lambda: defaultdict(lambda: 0))
        _RELOAD_COUNT = defaultdict(lambda: defaultdict(lambda: 0))
        _READ_COUNT = defaultdict(lambda: defaultdict(lambda: 0))
        _TO_SKIP = defaultdict(lambda: defaultdict(lambda: 0))
        _TO_FAIL = defaultdict(lambda: defaultdict(lambda: 0))

        _EVENTS = defaultdict(lambda: defaultdict(lambda: []))

        @classmethod
        def set_skip(cls, agent, key, skip):
            cls._TO_SKIP[agent][key] = skip

        @classmethod
        def set_fail(cls, agent, key, failcount):
            cls._TO_FAIL[agent][key] = failcount

        @classmethod
        def skip(cls, agent, key):
            doskip = cls._TO_SKIP[agent][key]
            if doskip == 0:
                return False
            cls._TO_SKIP[agent][key] -= 1
            return True

        @classmethod
        def fail(cls, agent, key):
            doskip = cls._TO_FAIL[agent][key]
            if doskip == 0:
                return False
            cls._TO_FAIL[agent][key] -= 1
            return True

        @classmethod
        def touch(cls, agent, key):
            cls._WRITE_COUNT[agent][key] += 1

        @classmethod
        def read(cls, agent, key):
            cls._READ_COUNT[agent][key] += 1

        @classmethod
        def set(cls, agent, key, value):
            cls._STATE[agent][key] = value

        @classmethod
        def get(cls, agent, key):
            if key in cls._STATE[agent]:
                return cls._STATE[agent][key]
            return None

        @classmethod
        def isset(cls, agent, key):
            return key in cls._STATE[agent]

        @classmethod
        def delete(cls, agent, key):
            if cls.isset(agent, key):
                del cls._STATE[agent][key]

        @classmethod
        def changecount(cls, agent, key):
            return cls._WRITE_COUNT[agent][key]

        @classmethod
        def readcount(cls, agent, key):
            return cls._READ_COUNT[agent][key]

        @classmethod
        def getevents(cls, agent, key):
            return cls._EVENTS[agent][key]

        @classmethod
        def reloadcount(cls, agent, key):
            return cls._RELOAD_COUNT[agent][key]

        @classmethod
        def reset(cls):
            cls._STATE = defaultdict(dict)
            cls._EVENTS = defaultdict(lambda: defaultdict(lambda: []))
            cls._WRITE_COUNT = defaultdict(lambda: defaultdict(lambda: 0))
            cls._READ_COUNT = defaultdict(lambda: defaultdict(lambda: 0))
            cls._TO_SKIP = defaultdict(lambda: defaultdict(lambda: 0))
            cls._RELOAD_COUNT = defaultdict(lambda: defaultdict(lambda: 0))

    @provider("test::Fail", name="test_fail")
    class Fail(ResourceHandler):

        def check_resource(self, ctx, resource):
            current = resource.clone()
            current.purged = not Provider.isset(resource.id.get_agent_name(), resource.key)

            if not current.purged:
                current.value = Provider.get(resource.id.get_agent_name(), resource.key)
            else:
                current.value = None

            return current

        def do_changes(self, ctx, resource, changes):
            raise Exception()

    @provider("test::FailFast", name="test_failfast")
    class FailFast(ResourceHandler):

        def check_resource(self, ctx, resource):
            raise Exception()

    @provider("test::Fact", name="test_fact")
    class Fact(ResourceHandler):

        def check_resource(self, ctx, resource):
            current = resource.clone()
            current.purged = not Provider.isset(resource.id.get_agent_name(), resource.key)

            current.value = "that"

            return current

        def do_changes(self, ctx, resource, changes):
            if resource.skip:
                raise SkipResource("can not deploy")
            if "purged" in changes:
                if changes["purged"]["desired"]:
                    Provider.delete(resource.id.get_agent_name(), resource.key)
                    ctx.set_purged()
                else:
                    Provider.set(resource.id.get_agent_name(), resource.key, "x")
                    ctx.set_created()
            else:
                ctx.set_updated()

        def facts(self, ctx: HandlerContext, resource: Resource) -> dict:
            if not Provider.isset(resource.id.get_agent_name(), resource.key):
                return {}
            elif resource.skipFact:
                raise SkipResource("Not ready")
            return {"fact": resource.factvalue}

    @provider("test::BadEvents", name="test_bad_events")
    class BadEvents(ResourceHandler):

        def check_resource(self, ctx, resource):
            current = resource.clone()
            return current

        def do_changes(self, ctx, resource, changes):
            pass

        def can_process_events(self) -> bool:
            return True

        def process_events(self, ctx, resource, events):
            raise Exception()

    waiter = Condition()

    @gen.coroutine
    def wait_for_done_with_waiters(client, env_id, version):
        # unhang waiters
        result = yield client.get_version(env_id, version)
        assert result.code == 200
        while (result.result["model"]["total"] - result.result["model"]["done"]) > 0:
            result = yield client.get_version(env_id, version)
            logger.info("waiting with waiters, %s resources done", result.result["model"]["done"])
            if result.result["model"]["done"] > 0:
                waiter.acquire()
                waiter.notifyAll()
                waiter.release()
            yield gen.sleep(0.1)

        return result

    @provider("test::Wait", name="test_wait")
    class Wait(ResourceHandler):

        def __init__(self, agent, io=None):
            super().__init__(agent, io)
            self.traceid = uuid.uuid4()

        def check_resource(self, ctx, resource):
            current = resource.clone()
            current.purged = not Provider.isset(resource.id.get_agent_name(), resource.key)

            if not current.purged:
                current.value = Provider.get(resource.id.get_agent_name(), resource.key)
            else:
                current.value = None

            return current

        def do_changes(self, ctx, resource, changes):
            logger.info("Hanging waiter %s", self.traceid)
            waiter.acquire()
            waiter.wait()
            waiter.release()
            logger.info("Releasing waiter %s", self.traceid)
            if "purged" in changes:
                if changes["purged"]["desired"]:
                    Provider.delete(resource.id.get_agent_name(), resource.key)
                    ctx.set_purged()
                else:
                    Provider.set(resource.id.get_agent_name(), resource.key, resource.value)
                    ctx.set_created()

            if "value" in changes:
                Provider.set(resource.id.get_agent_name(), resource.key, resource.value)
                ctx.set_updated()

    return ResourceContainer(Provider=Provider, wait_for_done_with_waiters=wait_for_done_with_waiters, waiter=waiter)


@pytest.mark.gen_test(timeout=150)
def test_dryrun_and_deploy(io_loop, server_multi, client_multi, resource_container):
    """
        dryrun and deploy a configuration model

        There is a second agent with an undefined resource. The server will shortcut the dryrun and deploy for this resource
        without an agent being present.
    """

    agentmanager = server_multi.get_endpoint(SLICE_AGENT_MANAGER)

    resource_container.Provider.reset()
    result = yield client_multi.create_project("env-test")
    project_id = result.result["project"]["id"]

    result = yield client_multi.create_environment(project_id=project_id, name="dev")
    env_id = result.result["environment"]["id"]

    agent = Agent(io_loop, hostname="node1", environment=env_id, agent_map={"agent1": "localhost"},
                  code_loader=False)
    agent.add_end_point_name("agent1")
    agent.start()

    yield retry_limited(lambda: len(agentmanager.sessions) == 1, 10)

    resource_container.Provider.set("agent1", "key2", "incorrect_value")
    resource_container.Provider.set("agent1", "key3", "value")

    version = int(time.time())

    resources = [{'key': 'key1',
                  'value': 'value1',
                  'id': 'test::Resource[agent1,key=key1],v=%d' % version,
                  'send_event': False,
                  'purged': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  'requires': ['test::Resource[agent1,key=key2],v=%d' % version],
                  },
                 {'key': 'key2',
                  'value': 'value2',
                  'id': 'test::Resource[agent1,key=key2],v=%d' % version,
                  'send_event': False,
                  'requires': [],
                  'purged': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  },
                 {'key': 'key3',
                  'value': None,
                  'id': 'test::Resource[agent1,key=key3],v=%d' % version,
                  'send_event': False,
                  'requires': [],
                  'purged': True,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  },
                 {'key': 'key4',
                  'value': execute.util.Unknown(source=None),
                  'id': 'test::Resource[agent2,key=key4],v=%d' % version,
                  'send_event': False,
                  'requires': [],
                  'purged': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  },
                 {'key': 'key5',
                  'value': "val",
                  'id': 'test::Resource[agent2,key=key5],v=%d' % version,
                  'send_event': False,
                  'requires': ['test::Resource[agent2,key=key4],v=%d' % version],
                  'purged': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  },
                 {'key': 'key6',
                  'value': "val",
                  'id': 'test::Resource[agent2,key=key6],v=%d' % version,
                  'send_event': False,
                  'requires': ['test::Resource[agent2,key=key5],v=%d' % version],
                  'purged': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  }
                 ]

    status = {'test::Resource[agent2,key=key4]': const.ResourceState.undefined}
    result = yield client_multi.put_version(tid=env_id, version=version, resources=resources, resource_state=status,
                                            unknowns=[], version_info={})
    assert result.code == 200

    mod_db = yield data.ConfigurationModel.get_version(uuid.UUID(env_id), version)
    undep = yield mod_db.get_undeployable()
    assert undep == ['test::Resource[agent2,key=key4]']

    undep = yield mod_db.get_skipped_for_undeployable()
    assert undep == ['test::Resource[agent2,key=key5]', 'test::Resource[agent2,key=key6]']

    # request a dryrun
    result = yield client_multi.dryrun_request(env_id, version)
    assert result.code == 200
    assert result.result["dryrun"]["total"] == len(resources)
    assert result.result["dryrun"]["todo"] == len(resources)

    # get the dryrun results
    result = yield client_multi.dryrun_list(env_id, version)
    assert result.code == 200
    assert len(result.result["dryruns"]) == 1

    while result.result["dryruns"][0]["todo"] > 0:
        result = yield client_multi.dryrun_list(env_id, version)
        yield gen.sleep(0.1)

    dry_run_id = result.result["dryruns"][0]["id"]
    result = yield client_multi.dryrun_report(env_id, dry_run_id)
    assert result.code == 200

    changes = result.result["dryrun"]["resources"]
    assert changes[resources[0]["id"]]["changes"]["purged"]["current"]
    assert not changes[resources[0]["id"]]["changes"]["purged"]["desired"]
    assert changes[resources[0]["id"]]["changes"]["value"]["current"] is None
    assert changes[resources[0]["id"]]["changes"]["value"]["desired"] == resources[0]["value"]

    assert changes[resources[1]["id"]]["changes"]["value"]["current"] == "incorrect_value"
    assert changes[resources[1]["id"]]["changes"]["value"]["desired"] == resources[1]["value"]

    assert not changes[resources[2]["id"]]["changes"]["purged"]["current"]
    assert changes[resources[2]["id"]]["changes"]["purged"]["desired"]

    # do a deploy
    result = yield client_multi.release_version(env_id, version, True)
    assert result.code == 200
    assert not result.result["model"]["deployed"]
    assert result.result["model"]["released"]
    assert result.result["model"]["total"] == 6
    assert result.result["model"]["result"] == "deploying"

    result = yield client_multi.get_version(env_id, version)
    assert result.code == 200

    while (result.result["model"]["total"] - result.result["model"]["done"]) > 0:
        result = yield client_multi.get_version(env_id, version)
        yield gen.sleep(0.1)

    assert result.result["model"]["done"] == len(resources)

    assert resource_container.Provider.isset("agent1", "key1")
    assert resource_container.Provider.get("agent1", "key1") == "value1"
    assert resource_container.Provider.get("agent1", "key2") == "value2"
    assert not resource_container.Provider.isset("agent1", "key3")

    actions = yield data.ResourceAction.get_list()
    assert sum([len(x.resource_version_ids) for x in actions if x.status == const.ResourceState.undefined]) == 1
    assert sum([len(x.resource_version_ids) for x in actions if x.status == const.ResourceState.skipped_for_undefined]) == 2

    agent.stop()


@pytest.mark.gen_test(timeout=100)
def test_deploy_with_undefined(io_loop, server_multi, client_multi, resource_container):
    """
         Test deploy of resource with undefined
    """

    # agent backoff makes this test unreliable or slow, so we turn it off
    backoff = inmanta.agent.agent.GET_RESOURCE_BACKOFF
    inmanta.agent.agent.GET_RESOURCE_BACKOFF = 0

    agentmanager = server_multi.get_endpoint(SLICE_AGENT_MANAGER)

    Config.set("config", "agent-interval", "100")

    resource_container.Provider.reset()
    result = yield client_multi.create_project("env-test")
    project_id = result.result["project"]["id"]

    result = yield client_multi.create_environment(project_id=project_id, name="dev")
    env_id = result.result["environment"]["id"]

    resource_container.Provider.set_skip("agent2", "key1", 1)

    agent = Agent(io_loop, hostname="node1", environment=env_id, agent_map={"agent1": "localhost", "agent2": "localhost"},
                  code_loader=False)
    agent.add_end_point_name("agent2")
    agent.start()

    yield retry_limited(lambda: len(agentmanager.sessions) == 1, 10)

    version = int(time.time())

    resources = [{'key': 'key1',
                  'value': 'value1',
                  'id': 'test::Resource[agent2,key=key1],v=%d' % version,
                  'send_event': False,
                  'purged': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  'requires': [],
                  },
                 {'key': 'key2',
                  'value': execute.util.Unknown(source=None),
                  'id': 'test::Resource[agent2,key=key2],v=%d' % version,
                  'send_event': False,
                  'purged': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  'requires': [],
                  },
                 {'key': 'key4',
                  'value': execute.util.Unknown(source=None),
                  'id': 'test::Resource[agent2,key=key4],v=%d' % version,
                  'send_event': False,
                  'requires': ['test::Resource[agent2,key=key1],v=%d' % version,
                               'test::Resource[agent2,key=key2],v=%d' % version],
                  'purged': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  },
                 {'key': 'key5',
                  'value': "val",
                  'id': 'test::Resource[agent2,key=key5],v=%d' % version,
                  'send_event': False,
                  'requires': ['test::Resource[agent2,key=key4],v=%d' % version],
                  'purged': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  }
                 ]

    status = {'test::Resource[agent2,key=key4]': const.ResourceState.undefined,
              'test::Resource[agent2,key=key2]': const.ResourceState.undefined}
    result = yield client_multi.put_version(tid=env_id, version=version, resources=resources, resource_state=status,
                                            unknowns=[], version_info={})
    assert result.code == 200

    # do a deploy
    result = yield client_multi.release_version(env_id, version, True)
    assert result.code == 200
    assert not result.result["model"]["deployed"]
    assert result.result["model"]["released"]
    assert result.result["model"]["total"] == len(resources)
    assert result.result["model"]["result"] == "deploying"

    # The server will mark the full version as deployed even though the agent has not done anything yet.
    result = yield client_multi.get_version(env_id, version)
    assert result.code == 200

    while (result.result["model"]["total"] - result.result["model"]["done"]) > 0:
        result = yield client_multi.get_version(env_id, version)
        yield gen.sleep(0.1)

    assert result.result["model"]["done"] == len(resources)
    assert result.code == 200

    actions = yield data.ResourceAction.get_list()
    assert len([x for x in actions if x.status == const.ResourceState.undefined]) >= 1

    result = yield client_multi.get_version(env_id, version)
    assert result.code == 200

    assert resource_container.Provider.changecount("agent2", "key4") == 0
    assert resource_container.Provider.changecount("agent2", "key5") == 0
    assert resource_container.Provider.changecount("agent2", "key1") == 0

    assert resource_container.Provider.readcount("agent2", "key4") == 0
    assert resource_container.Provider.readcount("agent2", "key5") == 0
    assert resource_container.Provider.readcount("agent2", "key1") == 1

    # Do a second deploy of the same model on agent2 with undefined resources
    yield agent.trigger_update("env_id", "agent2")

    result = yield client_multi.get_version(env_id, version, include_logs=True)
    import pprint
    pprint.pprint(result.result)

    def done():
        return resource_container.Provider.changecount("agent2", "key4") == 0 and \
            resource_container.Provider.changecount("agent2", "key5") == 0 and \
            resource_container.Provider.changecount("agent2", "key1") == 1 and \
            resource_container.Provider.readcount("agent2", "key4") == 0 and \
            resource_container.Provider.readcount("agent2", "key5") == 0 and \
            resource_container.Provider.readcount("agent2", "key1") == 2

    yield retry_limited(done, 100)

    agent.stop()
    inmanta.agent.agent.GET_RESOURCE_BACKOFF = backoff


@pytest.mark.gen_test(timeout=30)
def test_server_restart(resource_container, io_loop, server, mongo_db, client):
    """
        dryrun and deploy a configuration model
    """
    agentmanager = server.get_endpoint(SLICE_AGENT_MANAGER)

    resource_container.Provider.reset()
    result = yield client.create_project("env-test")
    project_id = result.result["project"]["id"]

    result = yield client.create_environment(project_id=project_id, name="dev")
    env_id = result.result["environment"]["id"]

    agent = Agent(io_loop, hostname="node1", environment=env_id, agent_map={"agent1": "localhost"},
                  code_loader=False)
    agent.add_end_point_name("agent1")
    agent.start()
    yield retry_limited(lambda: len(agentmanager.sessions) == 1, 10)

    resource_container.Provider.set("agent1", "key2", "incorrect_value")
    resource_container.Provider.set("agent1", "key3", "value")

    server.stop()

    ibl = InmantaBootloader()
    server = ibl.restserver
    ibl.start()
    agentmanager = server.get_endpoint(SLICE_AGENT_MANAGER)

    yield retry_limited(lambda: len(agentmanager.sessions) == 1, 10)

    version = int(time.time())

    resources = [{'key': 'key1',
                  'value': 'value1',
                  'id': 'test::Resource[agent1,key=key1],v=%d' % version,
                  'purged': False,
                  'send_event': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  'requires': ['test::Resource[agent1,key=key2],v=%d' % version],
                  },
                 {'key': 'key2',
                  'value': 'value2',
                  'id': 'test::Resource[agent1,key=key2],v=%d' % version,
                  'requires': [],
                  'purged': False,
                  'send_event': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  },
                 {'key': 'key3',
                  'value': None,
                  'id': 'test::Resource[agent1,key=key3],v=%d' % version,
                  'requires': [],
                  'purged': True,
                  'send_event': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  }
                 ]

    result = yield client.put_version(tid=env_id, version=version, resources=resources, unknowns=[], version_info={})
    assert result.code == 200

    # request a dryrun
    result = yield client.dryrun_request(env_id, version)
    assert result.code == 200
    assert result.result["dryrun"]["total"] == len(resources)
    assert result.result["dryrun"]["todo"] == len(resources)

    # get the dryrun results
    result = yield client.dryrun_list(env_id, version)
    assert result.code == 200
    assert len(result.result["dryruns"]) == 1

    while result.result["dryruns"][0]["todo"] > 0:
        result = yield client.dryrun_list(env_id, version)
        yield gen.sleep(0.1)

    dry_run_id = result.result["dryruns"][0]["id"]
    result = yield client.dryrun_report(env_id, dry_run_id)
    assert result.code == 200

    changes = result.result["dryrun"]["resources"]
    assert changes[resources[0]["id"]]["changes"]["purged"]["current"]
    assert not changes[resources[0]["id"]]["changes"]["purged"]["desired"]
    assert changes[resources[0]["id"]]["changes"]["value"]["current"] is None
    assert changes[resources[0]["id"]]["changes"]["value"]["desired"] == resources[0]["value"]

    assert changes[resources[1]["id"]]["changes"]["value"]["current"] == "incorrect_value"
    assert changes[resources[1]["id"]]["changes"]["value"]["desired"] == resources[1]["value"]

    assert not changes[resources[2]["id"]]["changes"]["purged"]["current"]
    assert changes[resources[2]["id"]]["changes"]["purged"]["desired"]

    # do a deploy
    result = yield client.release_version(env_id, version, True)
    assert result.code == 200
    assert not result.result["model"]["deployed"]
    assert result.result["model"]["released"]
    assert result.result["model"]["total"] == 3
    assert result.result["model"]["result"] == "deploying"

    result = yield client.get_version(env_id, version)
    assert result.code == 200

    while (result.result["model"]["total"] - result.result["model"]["done"]) > 0:
        result = yield client.get_version(env_id, version)
        yield gen.sleep(0.1)

    assert result.result["model"]["done"] == len(resources)

    assert resource_container.Provider.isset("agent1", "key1")
    assert resource_container.Provider.get("agent1", "key1") == "value1"
    assert resource_container.Provider.get("agent1", "key2") == "value2"
    assert not resource_container.Provider.isset("agent1", "key3")

    agent.stop()
    ibl.stop()


@pytest.mark.gen_test(timeout=30)
def test_spontaneous_deploy(resource_container, io_loop, server, client):
    """
        dryrun and deploy a configuration model
    """
    agentmanager = server.get_endpoint(SLICE_AGENT_MANAGER)

    resource_container.Provider.reset()
    result = yield client.create_project("env-test")
    project_id = result.result["project"]["id"]

    result = yield client.create_environment(project_id=project_id, name="dev")
    env_id = result.result["environment"]["id"]

    Config.set("config", "agent-interval", "2")
    Config.set("config", "agent-splay", "2")

    agent = Agent(io_loop, hostname="node1", environment=env_id, agent_map={"agent1": "localhost"},
                  code_loader=False)
    agent.add_end_point_name("agent1")
    agent.start()
    yield retry_limited(lambda: len(agentmanager.sessions) == 1, 10)

    resource_container.Provider.set("agent1", "key2", "incorrect_value")
    resource_container.Provider.set("agent1", "key3", "value")

    version = int(time.time())

    resources = [{'key': 'key1',
                  'value': 'value1',
                  'id': 'test::Resource[agent1,key=key1],v=%d' % version,
                  'purged': False,
                  'send_event': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  'requires': ['test::Resource[agent1,key=key2],v=%d' % version],
                  },
                 {'key': 'key2',
                  'value': 'value2',
                  'id': 'test::Resource[agent1,key=key2],v=%d' % version,
                  'requires': [],
                  'purged': False,
                  'send_event': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  },
                 {'key': 'key3',
                  'value': None,
                  'id': 'test::Resource[agent1,key=key3],v=%d' % version,
                  'requires': [],
                  'purged': True,
                  'send_event': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  }
                 ]

    result = yield client.put_version(tid=env_id, version=version, resources=resources, unknowns=[], version_info={})
    assert result.code == 200

    # do a deploy
    result = yield client.release_version(env_id, version, False)
    assert result.code == 200
    assert not result.result["model"]["deployed"]
    assert result.result["model"]["released"]
    assert result.result["model"]["total"] == 3
    assert result.result["model"]["result"] == "deploying"

    result = yield client.get_version(env_id, version)
    assert result.code == 200

    while (result.result["model"]["total"] - result.result["model"]["done"]) > 0:
        result = yield client.get_version(env_id, version)
        yield gen.sleep(0.1)

    assert result.result["model"]["done"] == len(resources)

    assert resource_container.Provider.isset("agent1", "key1")
    assert resource_container.Provider.get("agent1", "key1") == "value1"
    assert resource_container.Provider.get("agent1", "key2") == "value2"
    assert not resource_container.Provider.isset("agent1", "key3")

    agent.stop()


@pytest.mark.gen_test(timeout=30)
def test_failing_deploy_no_handler(resource_container, io_loop, server, client):
    """
        dryrun and deploy a configuration model
    """
    agentmanager = server.get_endpoint(SLICE_AGENT_MANAGER)

    resource_container.Provider.reset()
    result = yield client.create_project("env-test")
    project_id = result.result["project"]["id"]
    result = yield client.create_environment(project_id=project_id, name="dev")
    env_id = result.result["environment"]["id"]

    agent = Agent(io_loop, hostname="node1", environment=env_id, agent_map={"agent1": "localhost"},
                  code_loader=False)
    agent.add_end_point_name("agent1")
    agent.start()
    yield retry_limited(lambda: len(agentmanager.sessions) == 1, 10)

    version = int(time.time())

    resources = [{'key': 'key1',
                  'value': 'value1',
                  'id': 'test::Noprov[agent1,key=key1],v=%d' % version,
                  'purged': False,
                  'send_event': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  'requires': [],
                  }
                 ]

    result = yield client.put_version(tid=env_id, version=version, resources=resources, unknowns=[], version_info={})
    assert result.code == 200

    # do a deploy
    result = yield client.release_version(env_id, version, True)
    assert result.code == 200
    assert result.result["model"]["total"] == 1

    result = yield client.get_version(env_id, version)
    assert result.code == 200

    while (result.result["model"]["total"] - result.result["model"]["done"]) > 0:
        result = yield client.get_version(env_id, version)
        yield gen.sleep(0.1)

    assert result.result["model"]["done"] == len(resources)

    result = yield client.get_version(env_id, version, include_logs=True)

    final_log = result.result["resources"][0]["actions"][0]["messages"][-1]
    assert "traceback" in final_log["kwargs"]

    agent.stop()


@pytest.mark.gen_test
def test_dual_agent(resource_container, io_loop, server, client, environment):
    """
        dryrun and deploy a configuration model
    """
    resource_container.Provider.reset()
    myagent = agent.Agent(io_loop, hostname="node1", environment=environment,
                          agent_map={"agent1": "localhost", "agent2": "localhost"},
                          code_loader=False)
    myagent.add_end_point_name("agent1")
    myagent.add_end_point_name("agent2")
    myagent.start()
    yield retry_limited(lambda: len(server.get_endpoint("session")._sessions) == 1, 10)

    resource_container.Provider.set("agent1", "key1", "incorrect_value")
    resource_container.Provider.set("agent2", "key1", "incorrect_value")

    version = int(time.time())

    resources = [{'key': 'key1',
                  'value': 'value1',
                  'id': 'test::Wait[agent1,key=key1],v=%d' % version,
                  'purged': False,
                  'send_event': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  'requires': []
                  },
                 {'key': 'key2',
                  'value': 'value1',
                  'id': 'test::Wait[agent1,key=key2],v=%d' % version,
                  'purged': False,
                  'send_event': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  'requires': ['test::Wait[agent1,key=key1],v=%d' % version]
                  },
                 {'key': 'key1',
                  'value': 'value2',
                  'id': 'test::Wait[agent2,key=key1],v=%d' % version,
                  'purged': False,
                  'send_event': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  'requires': []
                  },
                 {'key': 'key2',
                  'value': 'value2',
                  'id': 'test::Wait[agent2,key=key2],v=%d' % version,
                  'purged': False,
                  'send_event': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  'requires': ['test::Wait[agent2,key=key1],v=%d' % version]
                  }]

    result = yield client.put_version(tid=environment, version=version, resources=resources, unknowns=[], version_info={})
    assert result.code == 200

    # expire rate limiting
    yield gen.sleep(0.5)
    # do a deploy
    result = yield client.release_version(environment, version, True)
    assert result.code == 200

    assert not result.result["model"]["deployed"]
    assert result.result["model"]["released"]
    assert result.result["model"]["total"] == 4

    result = yield client.get_version(environment, version)
    assert result.code == 200

    while (result.result["model"]["total"] - result.result["model"]["done"]) > 0:
        result = yield client.get_version(environment, version)
        resource_container.waiter.acquire()
        resource_container.waiter.notifyAll()
        resource_container.waiter.release()
        yield gen.sleep(0.1)

    assert result.result["model"]["done"] == len(resources)
    assert result.result["model"]["result"] == const.VersionState.success.name

    assert resource_container.Provider.isset("agent1", "key1")
    assert resource_container.Provider.get("agent1", "key1") == "value1"
    assert resource_container.Provider.get("agent2", "key1") == "value2"
    assert resource_container.Provider.get("agent1", "key2") == "value1"
    assert resource_container.Provider.get("agent2", "key2") == "value2"

    myagent.stop()


@pytest.mark.gen_test(timeout=60)
def test_snapshot_restore(resource_container, client, server, io_loop):
    """
        create a snapshot and restore it again
    """
    resource_container.Provider.reset()
    result = yield client.create_project("env-test")
    project_id = result.result["project"]["id"]

    result = yield client.create_environment(project_id=project_id, name="dev")
    env_id = result.result["environment"]["id"]

    agent = Agent(io_loop, hostname="node1", environment=env_id, agent_map={"agent1": "localhost"},
                  code_loader=False)
    agent.add_end_point_name("agent1")
    agent.start()
    yield retry_limited(lambda: len(server.get_endpoint("session")._sessions) == 1, 10)

    resource_container.Provider.set("agent1", "key", "value")

    version = int(time.time())

    resources = [{'key': 'key',
                  'value': 'value',
                  'id': 'test::Resource[agent1,key=key],v=%d' % version,
                  'requires': [],
                  'purged': False,
                  'send_event': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  },
                 {'key': 'key2',
                  'value': 'value',
                  'id': 'test::Resource[agent1,key=key2],v=%d' % version,
                  'requires': [],
                  'purged': False,
                  'send_event': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  }]

    result = yield client.put_version(tid=env_id, version=version, resources=resources, unknowns=[], version_info={})
    assert result.code == 200

    # deploy and wait until done
    result = yield client.release_version(env_id, version, True)
    assert result.code == 200

    result = yield client.get_version(env_id, version)
    assert result.code == 200
    while (result.result["model"]["total"] - result.result["model"]["done"]) > 0:
        result = yield client.get_version(env_id, version)
        yield gen.sleep(0.1)

    assert result.result["model"]["done"] == len(resources)

    # create a snapshot
    result = yield client.create_snapshot(env_id, "snap1")
    assert result.code == 200
    snapshot_id = result.result["snapshot"]["id"]

    result = yield client.list_snapshots(env_id)
    assert result.code == 200
    assert len(result.result["snapshots"]) == 1
    assert result.result["snapshots"][0]["id"] == snapshot_id

    while result.result["snapshots"][0]["finished"] is None:
        result = yield client.list_snapshots(env_id)
        assert result.code == 200
        yield gen.sleep(0.1)

    # Change the value of the resource
    resource_container.Provider.set("agent1", "key", "other")

    # try to do a restore
    result = yield client.restore_snapshot(env_id, snapshot_id)
    assert result.code == 200
    restore_id = result.result["restore"]["id"]

    result = yield client.list_restores(env_id)
    assert result.code == 200
    assert len(result.result["restores"]) == 1

    result = yield client.get_restore_status(env_id, restore_id)
    assert result.code == 200
    while result.result["restore"]["finished"] is None:
        result = yield client.get_restore_status(env_id, restore_id)
        assert result.code == 200
        yield gen.sleep(0.1)

    assert resource_container.Provider.get("agent1", "key") == "value"

    # get a snapshot
    result = yield client.get_snapshot(env_id, snapshot_id)
    assert result.code == 200
    assert result.result["snapshot"]["id"] == snapshot_id

    # delete the restore
    result = yield client.delete_restore(env_id, restore_id)
    assert result.code == 200

    # delete the snapshot
    result = yield client.delete_snapshot(env_id, snapshot_id)
    assert result.code == 200


@pytest.mark.gen_test
def test_server_agent_api(resource_container, client, server, io_loop):
    agentmanager = server.get_endpoint(SLICE_AGENT_MANAGER)

    result = yield client.create_project("env-test")
    project_id = result.result["project"]["id"]

    result = yield client.create_environment(project_id=project_id, name="dev")
    env_id = result.result["environment"]["id"]
    agent = Agent(io_loop, environment=env_id, hostname="agent1", agent_map={"agent1": "localhost"},
                  code_loader=False)
    agent.start()

    agent = Agent(io_loop, environment=env_id, hostname="agent2", agent_map={"agent2": "localhost"},
                  code_loader=False)
    agent.start()

    yield retry_limited(lambda: len(agentmanager.sessions) == 2, 10)
    assert len(agentmanager.sessions) == 2

    result = yield client.list_agent_processes(env_id)
    assert result.code == 200

    while len(result.result["processes"]) != 2:
        result = yield client.list_agent_processes(env_id)
        assert result.code == 200
        yield gen.sleep(0.1)

    assert len(result.result["processes"]) == 2
    agents = ["agent1", "agent2"]
    for proc in result.result["processes"]:
        assert proc["environment"] == env_id
        assert len(proc["endpoints"]) == 1
        assert proc["endpoints"][0]["name"] in agents
        agents.remove(proc["endpoints"][0]["name"])

    assert_equal_ish({'processes': [{'expired': None, 'environment': env_id,
                                     'endpoints': [{'name': UNKWN, 'process': UNKWN, 'id': UNKWN}], 'id': UNKWN,
                                     'hostname': UNKWN, 'first_seen': UNKWN, 'last_seen': UNKWN},
                                    {'expired': None, 'environment': env_id,
                                     'endpoints': [{'name': UNKWN, 'process': UNKWN, 'id': UNKWN}],
                                     'id': UNKWN, 'hostname': UNKWN, 'first_seen': UNKWN, 'last_seen': UNKWN}
                                    ]},
                     result.result, ['name', 'first_seen'])

    agentid = result.result["processes"][0]["id"]
    endpointid = [x["endpoints"][0]["id"] for x in result.result["processes"] if x["endpoints"][0]["name"] == "agent1"][0]

    result = yield client.get_agent_process(id=agentid)
    assert result.code == 200

    result = yield client.get_agent_process(id=uuid.uuid4())
    assert result.code == 404

    version = int(time.time())

    resources = [{'key': 'key',
                  'value': 'value',
                  'id': 'test::Resource[agent1,key=key],v=%d' % version,
                  'requires': [],
                  'purged': False,
                  'send_event': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  },
                 {'key': 'key2',
                  'value': 'value',
                  'id': 'test::Resource[agent1,key=key2],v=%d' % version,
                  'requires': [],
                  'purged': False,
                  'send_event': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  }]

    result = yield client.put_version(tid=env_id, version=version, resources=resources, unknowns=[], version_info={})
    assert result.code == 200

    result = yield client.list_agents(tid=env_id)
    assert result.code == 200

    shouldbe = {'agents': [
        {'last_failover': UNKWN, 'environment': env_id, 'paused': False,
         'primary': endpointid, 'name': 'agent1', 'state': 'up'}]}

    assert_equal_ish(shouldbe, result.result)

    result = yield client.list_agents(tid=uuid.uuid4())
    assert result.code == 404


@pytest.mark.gen_test
def test_get_facts(resource_container, client, server, io_loop):
    """
        Test retrieving facts from the agent
    """
    resource_container.Provider.reset()
    result = yield client.create_project("env-test")
    project_id = result.result["project"]["id"]

    result = yield client.create_environment(project_id=project_id, name="dev")
    env_id = result.result["environment"]["id"]

    agent = Agent(io_loop, hostname="node1", environment=env_id, agent_map={"agent1": "localhost"},
                  code_loader=False)
    agent.add_end_point_name("agent1")
    agent.start()
    yield retry_limited(lambda: len(server.get_endpoint("session")._sessions) == 1, 10)

    resource_container.Provider.set("agent1", "key", "value")

    version = int(time.time())

    resource_id_wov = "test::Resource[agent1,key=key]"
    resource_id = "%s,v=%d" % (resource_id_wov, version)

    resources = [{'key': 'key',
                  'value': 'value',
                  'id': resource_id,
                  'requires': [],
                  'purged': False,
                  'send_event': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  }]

    result = yield client.put_version(tid=env_id, version=version, resources=resources, unknowns=[], version_info={})
    assert result.code == 200
    result = yield client.release_version(env_id, version, True)
    assert result.code == 200

    result = yield client.get_param(env_id, "length", resource_id_wov)
    assert result.code == 503

    env_uuid = uuid.UUID(env_id)
    params = yield data.Parameter.get_list(environment=env_uuid, resource_id=resource_id_wov)
    while len(params) < 3:
        params = yield data.Parameter.get_list(environment=env_uuid, resource_id=resource_id_wov)
        yield gen.sleep(0.1)

    result = yield client.get_param(env_id, "key1", resource_id_wov)
    assert result.code == 200


@pytest.mark.gen_test
def test_purged_facts(resource_container, client, server, io_loop, environment):
    """
        Test if facts are purged when the resource is purged.
    """
    resource_container.Provider.reset()
    agent = Agent(io_loop, hostname="node1", environment=environment, agent_map={"agent1": "localhost"},
                  code_loader=False)
    agent.add_end_point_name("agent1")
    agent.start()
    yield retry_limited(lambda: len(server.get_endpoint("session")._sessions) == 1, 10)

    resource_container.Provider.set("agent1", "key", "value")

    version = 1
    resource_id_wov = "test::Resource[agent1,key=key]"
    resource_id = "%s,v=%d" % (resource_id_wov, version)

    resources = [{'key': 'key',
                  'value': 'value',
                  'id': resource_id,
                  'requires': [],
                  'purged': False,
                  'send_event': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  }]

    result = yield client.put_version(tid=environment, version=version, resources=resources, unknowns=[], version_info={})
    assert result.code == 200
    result = yield client.release_version(environment, version, True)
    assert result.code == 200

    result = yield client.get_param(environment, "length", resource_id_wov)
    assert result.code == 503

    env_uuid = uuid.UUID(environment)
    params = yield data.Parameter.get_list(environment=env_uuid, resource_id=resource_id_wov)
    while len(params) < 3:
        params = yield data.Parameter.get_list(environment=env_uuid, resource_id=resource_id_wov)
        yield gen.sleep(0.1)

    result = yield client.get_param(environment, "key1", resource_id_wov)
    assert result.code == 200

    # Purge the resource
    version = 2
    resources[0]["id"] = "%s,v=%d" % (resource_id_wov, version)
    resources[0]["purged"] = True
    result = yield client.put_version(tid=environment, version=version, resources=resources, unknowns=[], version_info={})
    assert result.code == 200
    result = yield client.release_version(environment, version, True)
    assert result.code == 200

    result = yield client.get_version(environment, version)
    assert result.code == 200
    while (result.result["model"]["total"] - result.result["model"]["done"]) > 0:
        result = yield client.get_version(environment, version)
        yield gen.sleep(0.1)

    assert result.result["model"]["done"] == len(resources)

    # The resource facts should be purged
    result = yield client.get_param(environment, "length", resource_id_wov)
    assert result.code == 503


@pytest.mark.gen_test
def test_get_facts_extended(io_loop, server, client, resource_container, environment):
    """
        dryrun and deploy a configuration model automatically
    """
    agentmanager = server.get_endpoint(SLICE_AGENT_MANAGER)
    # allow very rapid fact refresh
    agentmanager._fact_resource_block = 0.1

    resource_container.Provider.reset()
    agent = Agent(io_loop, hostname="node1", environment=environment, agent_map={"agent1": "localhost"},
                  code_loader=False)
    agent.add_end_point_name("agent1")
    agent.start()
    yield retry_limited(lambda: len(agentmanager.sessions) == 1, 10)

    version = int(time.time())

    # mark some as existing
    resource_container.Provider.set("agent1", "key1", "value")
    resource_container.Provider.set("agent1", "key2", "value")
    resource_container.Provider.set("agent1", "key4", "value")
    resource_container.Provider.set("agent1", "key5", "value")
    resource_container.Provider.set("agent1", "key6", "value")
    resource_container.Provider.set("agent1", "key7", "value")

    resources = [{'key': 'key1',
                  'value': 'value1',
                  'id': 'test::Fact[agent1,key=key1],v=%d' % version,
                  'send_event': False,
                  'purged': False,
                  'skip': True,
                  'skipFact': False,
                  'factvalue': "fk1",
                  'requires': [],
                  },
                 {'key': 'key2',
                  'value': 'value1',
                  'id': 'test::Fact[agent1,key=key2],v=%d' % version,
                  'send_event': False,
                  'purged': False,
                  'skip': False,
                  'skipFact': False,
                  'factvalue': "fk2",
                  'requires': [],
                  },
                 {'key': 'key3',
                  'value': 'value1',
                  'id': 'test::Fact[agent1,key=key3],v=%d' % version,
                  'send_event': False,
                  'purged': False,
                  'skip': False,
                  'skipFact': False,
                  'factvalue': "fk3",
                  'requires': [],
                  },
                 {'key': 'key4',
                  'value': 'value1',
                  'id': 'test::Fact[agent1,key=key4],v=%d' % version,
                  'send_event': False,
                  'purged': False,
                  'skip': False,
                  'skipFact': False,
                  'factvalue': "fk4",
                  'requires': [],
                  },
                 {'key': 'key5',
                  'value': 'value1',
                  'id': 'test::Fact[agent1,key=key5],v=%d' % version,
                  'send_event': False,
                  'purged': False,
                  'skip': False,
                  'skipFact': True,
                  'factvalue': None,
                  'requires': [],
                  },
                 {'key': 'key6',
                  'value': 'value1',
                  'id': 'test::Fact[agent1,key=key6],v=%d' % version,
                  'send_event': False,
                  'purged': False,
                  'skip': False,
                  'skipFact': False,
                  'factvalue': None,
                  'requires': [],
                  },
                 {'key': 'key7',
                  'value': 'value1',
                  'id': 'test::Fact[agent1,key=key7],v=%d' % version,
                  'send_event': False,
                  'purged': False,
                  'skip': False,
                  'skipFact': False,
                  'factvalue': "",
                  'requires': [],
                  },
                 ]

    resource_states = {'test::Fact[agent1,key=key4],v=%d' % version: const.ResourceState.undefined,
                       'test::Fact[agent1,key=key5],v=%d' % version: const.ResourceState.undefined}

    @gen.coroutine
    def get_fact(rid, result_code=200, limit=10, lower_limit=2):
        lower_limit = limit - lower_limit
        result = yield client.get_param(environment, "fact", rid)

        # add minimal nr of reps or failure cases
        while (result.code != result_code and limit > 0) or limit > lower_limit:
            limit -= 1
            yield gen.sleep(0.1)
            result = yield client.get_param(environment, "fact", rid)

        assert result.code == result_code
        return result

    result = yield client.put_version(tid=environment,
                                      version=version,
                                      resources=resources,
                                      unknowns=[],
                                      version_info={},
                                      resource_state=resource_states)
    assert result.code == 200

    yield get_fact('test::Fact[agent1,key=key1]')  # undeployable
    yield get_fact('test::Fact[agent1,key=key2]')  # normal
    yield get_fact('test::Fact[agent1,key=key3]', 503)  # not present
    yield get_fact('test::Fact[agent1,key=key4]')  # unknown
    yield get_fact('test::Fact[agent1,key=key5]', 503)  # broken
    f6 = yield get_fact('test::Fact[agent1,key=key6]')  # normal
    f7 = yield get_fact('test::Fact[agent1,key=key7]')  # normal

    assert f6.result["parameter"]["value"] == 'None'
    assert f7.result["parameter"]["value"] == ""

    result = yield client.release_version(environment, version, True)
    assert result.code == 200

    result = yield client.get_version(environment, version)
    while (result.result["model"]["total"] - result.result["model"]["done"]) > 0:
        result = yield client.get_version(environment, version)
        yield gen.sleep(0.1)

    yield get_fact('test::Fact[agent1,key=key1]')  # undeployable
    yield get_fact('test::Fact[agent1,key=key2]')  # normal
    yield get_fact('test::Fact[agent1,key=key3]')  # not present -> present
    yield get_fact('test::Fact[agent1,key=key4]')  # unknown
    yield get_fact('test::Fact[agent1,key=key5]', 503)  # broken

    agent.stop()


@pytest.mark.gen_test
def test_get_set_param(resource_container, client, server, io_loop):
    """
        Test getting and setting params
    """
    resource_container.Provider.reset()
    result = yield client.create_project("env-test")
    project_id = result.result["project"]["id"]

    result = yield client.create_environment(project_id=project_id, name="dev")
    env_id = result.result["environment"]["id"]

    result = yield client.set_param(tid=env_id, id="key10", value="value10", source="user")
    assert result.code == 200

    result = yield client.get_param(tid=env_id, id="key10")
    assert result.code == 200
    assert result.result["parameter"]["value"] == "value10"

    result = yield client.delete_param(tid=env_id, id="key10")
    assert result.code == 200


@pytest.mark.gen_test
def test_unkown_parameters(resource_container, client, server, io_loop):
    """
        Test retrieving facts from the agent
    """
    resource_container.Provider.reset()
    result = yield client.create_project("env-test")
    project_id = result.result["project"]["id"]

    result = yield client.create_environment(project_id=project_id, name="dev")
    env_id = result.result["environment"]["id"]

    agent = Agent(io_loop, hostname="node1", environment=env_id, agent_map={"agent1": "localhost"},
                  code_loader=False)
    agent.add_end_point_name("agent1")
    agent.start()
    yield retry_limited(lambda: len(server.get_endpoint("session")._sessions) == 1, 10)

    resource_container.Provider.set("agent1", "key", "value")

    version = int(time.time())

    resource_id_wov = "test::Resource[agent1,key=key]"
    resource_id = "%s,v=%d" % (resource_id_wov, version)

    resources = [{'key': 'key',
                  'value': 'value',
                  'id': resource_id,
                  'requires': [],
                  'purged': False,
                  'send_event': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  }]

    unknowns = [{"resource": resource_id_wov, "parameter": "length", "source": "fact"}]
    result = yield client.put_version(tid=env_id, version=version, resources=resources, unknowns=unknowns,
                                      version_info={})
    assert result.code == 200

    result = yield client.release_version(env_id, version, True)
    assert result.code == 200

    yield server.get_endpoint("server").renew_expired_facts()

    env_id = uuid.UUID(env_id)
    params = yield data.Parameter.get_list(environment=env_id, resource_id=resource_id_wov)
    while len(params) < 3:
        params = yield data.Parameter.get_list(environment=env_id, resource_id=resource_id_wov)
        yield gen.sleep(0.1)

    result = yield client.get_param(env_id, "length", resource_id_wov)
    assert result.code == 200


@pytest.mark.gen_test()
def test_fail(resource_container, client, server, io_loop):
    """
        Test results when a step fails
    """
    resource_container.Provider.reset()
    result = yield client.create_project("env-test")
    project_id = result.result["project"]["id"]

    result = yield client.create_environment(project_id=project_id, name="dev")
    env_id = result.result["environment"]["id"]

    agent = Agent(io_loop, hostname="node1", environment=env_id, agent_map={"agent1": "localhost"},
                  code_loader=False, poolsize=10)
    agent.add_end_point_name("agent1")
    agent.start()
    yield retry_limited(lambda: len(server.get_endpoint("session")._sessions) == 1, 10)

    resource_container.Provider.set("agent1", "key", "value")

    version = int(time.time())

    resources = [{'key': 'key',
                  'value': 'value',
                  'id': 'test::Fail[agent1,key=key],v=%d' % version,
                  'requires': [],
                  'purged': False,
                  'send_event': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  },
                 {'key': 'key2',
                  'value': 'value',
                  'id': 'test::Resource[agent1,key=key2],v=%d' % version,
                  'requires': ['test::Fail[agent1,key=key],v=%d' % version],
                  'purged': False,
                  'send_event': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  },
                 {'key': 'key3',
                  'value': 'value',
                  'id': 'test::Resource[agent1,key=key3],v=%d' % version,
                  'requires': ['test::Fail[agent1,key=key],v=%d' % version],
                  'purged': False,
                  'send_event': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  },
                 {'key': 'key4',
                  'value': 'value',
                  'id': 'test::Resource[agent1,key=key4],v=%d' % version,
                  'requires': ['test::Resource[agent1,key=key3],v=%d' % version],
                  'purged': False,
                  'send_event': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  },
                 {'key': 'key5',
                  'value': 'value',
                  'id': 'test::Resource[agent1,key=key5],v=%d' % version,
                  'requires': ['test::Resource[agent1,key=key4],v=%d' % version,
                               'test::Fail[agent1,key=key],v=%d' % version],
                  'purged': False,
                  'send_event': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  }]

    result = yield client.put_version(tid=env_id, version=version, resources=resources, unknowns=[], version_info={})
    assert result.code == 200

    # deploy and wait until done
    result = yield client.release_version(env_id, version, True)
    assert result.code == 200

    result = yield client.get_version(env_id, version)
    assert result.code == 200
    while (result.result["model"]["total"] - result.result["model"]["done"]) > 0:
        result = yield client.get_version(env_id, version)
        yield gen.sleep(0.1)

    assert result.result["model"]["done"] == len(resources)

    states = {x["id"]: x["status"] for x in result.result["resources"]}

    assert states['test::Fail[agent1,key=key],v=%d' % version] == "failed"
    assert states['test::Resource[agent1,key=key2],v=%d' % version] == "skipped"
    assert states['test::Resource[agent1,key=key3],v=%d' % version] == "skipped"
    assert states['test::Resource[agent1,key=key4],v=%d' % version] == "skipped"
    assert states['test::Resource[agent1,key=key5],v=%d' % version] == "skipped"


@pytest.mark.gen_test(timeout=15)
def test_wait(resource_container, client, server, io_loop):
    """
        If this test fail due to timeout,
        this is probably due to the mechanism in the agent that prevents pulling resources in very rapid succession.

        If the test server is slow, a get_resources call takes a long time,
        this makes the back-off longer

        this test deploys two models in rapid successions, if the server is slow, this may fail due to the back-off
    """
    resource_container.Provider.reset()

    # setup project
    result = yield client.create_project("env-test")
    project_id = result.result["project"]["id"]

    # setup env
    result = yield client.create_environment(project_id=project_id, name="dev")
    env_id = result.result["environment"]["id"]

    # setup agent
    agent = Agent(io_loop, hostname="node1", environment=env_id, agent_map={"agent1": "localhost"},
                  code_loader=False, poolsize=10)
    agent.add_end_point_name("agent1")
    agent.start()

    # wait for agent
    yield retry_limited(lambda: len(server.get_endpoint("session")._sessions) == 1, 10)

    # set the deploy environment
    resource_container.Provider.set("agent1", "key", "value")

    def make_version(offset=0):
        version = int(time.time() + offset)

        resources = [{'key': 'key',
                      'value': 'value',
                      'id': 'test::Wait[agent1,key=key],v=%d' % version,
                      'requires': [],
                      'purged': False,
                      'send_event': False,
                      'state_id': '',
                      'allow_restore': True,
                      'allow_snapshot': True,
                      },
                     {'key': 'key2',
                      'value': 'value',
                      'id': 'test::Resource[agent1,key=key2],v=%d' % version,
                      'requires': ['test::Wait[agent1,key=key],v=%d' % version],
                      'purged': False,
                      'send_event': False,
                      'state_id': '',
                      'allow_restore': True,
                      'allow_snapshot': True,
                      },
                     {'key': 'key3',
                      'value': 'value',
                      'id': 'test::Resource[agent1,key=key3],v=%d' % version,
                      'requires': [],
                      'purged': False,
                      'send_event': False,
                      'state_id': '',
                      'allow_restore': True,
                      'allow_snapshot': True,
                      },
                     {'key': 'key4',
                      'value': 'value',
                      'id': 'test::Resource[agent1,key=key4],v=%d' % version,
                      'requires': ['test::Resource[agent1,key=key3],v=%d' % version],
                      'purged': False,
                      'send_event': False,
                      'state_id': '',
                      'allow_restore': True,
                      'allow_snapshot': True,
                      },
                     {'key': 'key5',
                      'value': 'value',
                      'id': 'test::Resource[agent1,key=key5],v=%d' % version,
                      'requires': ['test::Resource[agent1,key=key4],v=%d' % version,
                                   'test::Wait[agent1,key=key],v=%d' % version],
                      'purged': False,
                      'send_event': False,
                      'state_id': '',
                      'allow_restore': True,
                      'allow_snapshot': True,
                      }]
        return version, resources

    @gen.coroutine
    def wait_for_resources(version, n):
        result = yield client.get_version(env_id, version)
        assert result.code == 200

        while result.result["model"]["done"] < n:
            result = yield client.get_version(env_id, version)
            yield gen.sleep(0.1)
        assert result.result["model"]["done"] == n

    logger.info("setup done")

    version1, resources = make_version()
    result = yield client.put_version(tid=env_id, version=version1, resources=resources, unknowns=[], version_info={})
    assert result.code == 200

    logger.info("first version pushed")

    # deploy and wait until one is ready
    result = yield client.release_version(env_id, version1, True)
    assert result.code == 200

    logger.info("first version released")

    yield wait_for_resources(version1, 2)

    logger.info("first version, 2 resources deployed")

    version2, resources = make_version(3)
    result = yield client.put_version(tid=env_id, version=version2, resources=resources, unknowns=[], version_info={})
    assert result.code == 200

    logger.info("second version pushed %f", time.time())

    yield gen.sleep(1)

    logger.info("wait to expire load limiting%f", time.time())

    # deploy and wait until done
    result = yield client.release_version(env_id, version2, True)
    assert result.code == 200

    logger.info("second version released")

    yield resource_container.wait_for_done_with_waiters(client, env_id, version2)

    logger.info("second version complete")

    result = yield client.get_version(env_id, version2)
    assert result.code == 200
    for x in result.result["resources"]:
        assert x["status"] == const.ResourceState.deployed.name

    result = yield client.get_version(env_id, version1)
    assert result.code == 200
    states = {x["id"]: x["status"] for x in result.result["resources"]}

    assert states['test::Wait[agent1,key=key],v=%d' % version1] == const.ResourceState.deployed.name
    assert states['test::Resource[agent1,key=key2],v=%d' % version1] == const.ResourceState.available.name
    assert states['test::Resource[agent1,key=key3],v=%d' % version1] == const.ResourceState.deployed.name
    assert states['test::Resource[agent1,key=key4],v=%d' % version1] == const.ResourceState.deployed.name
    assert states['test::Resource[agent1,key=key5],v=%d' % version1] == const.ResourceState.available.name


@pytest.mark.gen_test(timeout=15)
def test_multi_instance(resource_container, client, server, io_loop):
    """
       Test for multi threaded deploy
    """
    resource_container.Provider.reset()

    # setup project
    result = yield client.create_project("env-test")
    project_id = result.result["project"]["id"]

    # setup env
    result = yield client.create_environment(project_id=project_id, name="dev")
    env_id = result.result["environment"]["id"]

    # setup agent
    agent = Agent(io_loop, hostname="node1", environment=env_id,
                  agent_map={"agent1": "localhost", "agent2": "localhost", "agent3": "localhost"},
                  code_loader=False, poolsize=1)
    agent.add_end_point_name("agent1")
    agent.add_end_point_name("agent2")
    agent.add_end_point_name("agent3")

    agent.start()

    # wait for agent
    yield retry_limited(lambda: len(server.get_endpoint("session")._sessions) == 1, 10)

    # set the deploy environment
    resource_container.Provider.set("agent1", "key", "value")
    resource_container.Provider.set("agent2", "key", "value")
    resource_container.Provider.set("agent3", "key", "value")

    def make_version(offset=0):
        version = int(time.time() + offset)
        resources = []
        for agent in ["agent1", "agent2", "agent3"]:
            resources.extend([{'key': 'key',
                               'value': 'value',
                               'id': 'test::Wait[%s,key=key],v=%d' % (agent, version),
                               'requires': ['test::Resource[%s,key=key3],v=%d' % (agent, version)],
                               'purged': False,
                               'send_event': False,
                               'state_id': '',
                               'allow_restore': True,
                               'allow_snapshot': True,
                               },
                              {'key': 'key2',
                               'value': 'value',
                               'id': 'test::Resource[%s,key=key2],v=%d' % (agent, version),
                               'requires': ['test::Wait[%s,key=key],v=%d' % (agent, version)],
                               'purged': False,
                               'send_event': False,
                               'state_id': '',
                               'allow_restore': True,
                               'allow_snapshot': True,
                               },
                              {'key': 'key3',
                               'value': 'value',
                               'id': 'test::Resource[%s,key=key3],v=%d' % (agent, version),
                               'requires': [],
                               'purged': False,
                               'send_event': False,
                               'state_id': '',
                               'allow_restore': True,
                               'allow_snapshot': True,
                               },
                              {'key': 'key4',
                               'value': 'value',
                               'id': 'test::Resource[%s,key=key4],v=%d' % (agent, version),
                               'requires': ['test::Resource[%s,key=key3],v=%d' % (agent, version)],
                               'purged': False,
                               'send_event': False,
                               'state_id': '',
                               'allow_restore': True,
                               'allow_snapshot': True,
                               },
                              {'key': 'key5',
                               'value': 'value',
                               'id': 'test::Resource[%s,key=key5],v=%d' % (agent, version),
                               'requires': ['test::Resource[%s,key=key4],v=%d' % (agent, version),
                                            'test::Wait[%s,key=key],v=%d' % (agent, version)],
                               'purged': False,
                               'send_event': False,
                               'state_id': '',
                               'allow_restore': True,
                               'allow_snapshot': True,
                               }])
        return version, resources

    @gen.coroutine
    def wait_for_resources(version, n):
        result = yield client.get_version(env_id, version)
        assert result.code == 200

        def done_per_agent(result):
            done = [x for x in result.result["resources"] if x["status"] == "deployed"]
            peragent = groupby(done, lambda x: x["agent"])
            return {agent: len([x for x in grp]) for agent, grp in peragent}

        def mindone(result):
            alllist = done_per_agent(result).values()
            if(len(alllist) == 0):
                return 0
            return min(alllist)

        while mindone(result) < n:
            yield gen.sleep(0.1)
            result = yield client.get_version(env_id, version)
        assert mindone(result) >= n

    logger.info("setup done")

    version1, resources = make_version()
    result = yield client.put_version(tid=env_id, version=version1, resources=resources, unknowns=[], version_info={})
    assert result.code == 200

    logger.info("first version pushed")

    # deploy and wait until one is ready
    result = yield client.release_version(env_id, version1, True)
    assert result.code == 200

    logger.info("first version released")
    # timeout on single thread!
    yield wait_for_resources(version1, 1)

    yield resource_container.wait_for_done_with_waiters(client, env_id, version1)

    logger.info("first version complete")


@pytest.mark.gen_test
def test_cross_agent_deps(resource_container, io_loop, server, client):
    """
        deploy a configuration model with cross host dependency
    """
    agentmanager = server.get_endpoint(SLICE_AGENT_MANAGER)

    resource_container.Provider.reset()
    # config for recovery mechanism
    Config.set("config", "agent-interval", "10")
    result = yield client.create_project("env-test")
    project_id = result.result["project"]["id"]

    result = yield client.create_environment(project_id=project_id, name="dev")
    env_id = result.result["environment"]["id"]

    agent = Agent(io_loop, hostname="node1", environment=env_id, agent_map={"agent1": "localhost"},
                  code_loader=False)
    agent.add_end_point_name("agent1")
    agent.start()
    yield retry_limited(lambda: len(agentmanager.sessions) == 1, 10)

    agent2 = Agent(io_loop, hostname="node2", environment=env_id, agent_map={"agent2": "localhost"},
                   code_loader=False)
    agent2.add_end_point_name("agent2")
    agent2.start()
    yield retry_limited(lambda: len(agentmanager.sessions) == 2, 10)

    resource_container.Provider.set("agent1", "key2", "incorrect_value")
    resource_container.Provider.set("agent1", "key3", "value")

    version = int(time.time())

    resources = [{'key': 'key1',
                  'value': 'value1',
                  'id': 'test::Resource[agent1,key=key1],v=%d' % version,
                  'purged': False,
                  'state_id': '',
                  'send_event': False,
                  'allow_restore': True,
                  'allow_snapshot': True,
                  'requires': ['test::Wait[agent1,key=key2],v=%d' % version, 'test::Resource[agent2,key=key3],v=%d' % version],
                  },
                 {'key': 'key2',
                  'value': 'value2',
                  'id': 'test::Wait[agent1,key=key2],v=%d' % version,
                  'requires': [],
                  'purged': False,
                  'state_id': '',
                  'send_event': False,
                  'allow_restore': True,
                  'allow_snapshot': True,
                  },
                 {'key': 'key3',
                  'value': 'value3',
                  'id': 'test::Resource[agent2,key=key3],v=%d' % version,
                  'requires': [],
                  'purged': False,
                  'state_id': '',
                  'send_event': False,
                  'allow_restore': True,
                  'allow_snapshot': True,
                  },
                 {'key': 'key4',
                  'value': 'value4',
                  'id': 'test::Resource[agent2,key=key4],v=%d' % version,
                  'requires': [],
                  'purged': False,
                  'state_id': '',
                  'send_event': False,
                  'allow_restore': True,
                  'allow_snapshot': True,
                  }
                 ]

    result = yield client.put_version(tid=env_id, version=version, resources=resources, unknowns=[], version_info={})
    assert result.code == 200

    # do a deploy
    result = yield client.release_version(env_id, version, True)
    assert result.code == 200
    assert not result.result["model"]["deployed"]
    assert result.result["model"]["released"]
    assert result.result["model"]["total"] == 4
    assert result.result["model"]["result"] == const.VersionState.deploying.name

    result = yield client.get_version(env_id, version)
    assert result.code == 200

    while result.result["model"]["done"] == 0:
        result = yield client.get_version(env_id, version)
        yield gen.sleep(0.1)

    result = yield resource_container.wait_for_done_with_waiters(client, env_id, version)

    assert result.result["model"]["done"] == len(resources)
    assert result.result["model"]["result"] == const.VersionState.success.name

    assert resource_container.Provider.isset("agent1", "key1")
    assert resource_container.Provider.get("agent1", "key1") == "value1"
    assert resource_container.Provider.get("agent1", "key2") == "value2"
    assert resource_container.Provider.get("agent2", "key3") == "value3"

    agent.stop()
    agent2.stop()


@pytest.mark.gen_test(timeout=30)
def test_dryrun_scale(resource_container, io_loop, server, client):
    """
        test dryrun scaling
    """
    agentmanager = server.get_endpoint(SLICE_AGENT_MANAGER)

    resource_container.Provider.reset()
    result = yield client.create_project("env-test")
    project_id = result.result["project"]["id"]

    result = yield client.create_environment(project_id=project_id, name="dev")
    env_id = result.result["environment"]["id"]

    agent = Agent(io_loop, hostname="node1", environment=env_id, agent_map={"agent1": "localhost"},
                  code_loader=False)
    agent.add_end_point_name("agent1")
    agent.start()
    yield retry_limited(lambda: len(agentmanager.sessions) == 1, 10)

    version = int(time.time())

    resources = []
    for i in range(1, 100):
        resources.append({'key': 'key%d' % i,
                          'value': 'value%d' % i,
                          'id': 'test::Resource[agent1,key=key%d],v=%d' % (i, version),
                          'purged': False,
                          'state_id': '',
                          'send_event': False,
                          'allow_restore': True,
                          'allow_snapshot': True,
                          'requires': [],
                          })

    result = yield client.put_version(tid=env_id, version=version, resources=resources, unknowns=[], version_info={})
    assert result.code == 200

    # request a dryrun
    result = yield client.dryrun_request(env_id, version)
    assert result.code == 200
    assert result.result["dryrun"]["total"] == len(resources)
    assert result.result["dryrun"]["todo"] == len(resources)

    # get the dryrun results
    result = yield client.dryrun_list(env_id, version)
    assert result.code == 200
    assert len(result.result["dryruns"]) == 1

    while result.result["dryruns"][0]["todo"] > 0:
        result = yield client.dryrun_list(env_id, version)
        yield gen.sleep(0.1)

    dry_run_id = result.result["dryruns"][0]["id"]
    result = yield client.dryrun_report(env_id, dry_run_id)
    assert result.code == 200

    agent.stop()


@pytest.mark.gen_test(timeout=30)
def test_dryrun_failures(resource_container, io_loop, server, client):
    """
        test dryrun scaling
    """
    agentmanager = server.get_endpoint(SLICE_AGENT_MANAGER)

    resource_container.Provider.reset()
    result = yield client.create_project("env-test")
    project_id = result.result["project"]["id"]

    result = yield client.create_environment(project_id=project_id, name="dev")
    env_id = result.result["environment"]["id"]

    agent = Agent(io_loop, hostname="node1", environment=env_id, agent_map={"agent1": "localhost"},
                  code_loader=False)
    agent.add_end_point_name("agent1")
    agent.start()
    yield retry_limited(lambda: len(agentmanager.sessions) == 1, 10)

    version = int(time.time())

    resources = [{'key': 'key1',
                  'value': 'value1',
                  'id': 'test::Noprov[agent1,key=key1],v=%d' % version,
                  'purged': False,
                  'send_event': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  'requires': [],
                  },
                 {'key': 'key2',
                  'value': 'value2',
                  'id': 'test::FailFast[agent1,key=key2],v=%d' % version,
                  'purged': False,
                  'send_event': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  'requires': [],
                  },
                 {'key': 'key2',
                  'value': 'value2',
                  'id': 'test::DoesNotExist[agent1,key=key2],v=%d' % version,
                  'purged': False,
                  'send_event': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  'requires': [],
                  }
                 ]

    result = yield client.put_version(tid=env_id, version=version, resources=resources, unknowns=[], version_info={})
    assert result.code == 200

    # request a dryrun
    result = yield client.dryrun_request(env_id, version)
    assert result.code == 200
    assert result.result["dryrun"]["total"] == len(resources)
    assert result.result["dryrun"]["todo"] == len(resources)

    # get the dryrun results
    result = yield client.dryrun_list(env_id, version)
    assert result.code == 200
    assert len(result.result["dryruns"]) == 1

    while result.result["dryruns"][0]["todo"] > 0:
        result = yield client.dryrun_list(env_id, version)
        print(result.result)
        yield gen.sleep(0.1)

    dry_run_id = result.result["dryruns"][0]["id"]
    result = yield client.dryrun_report(env_id, dry_run_id)
    assert result.code == 200

    resources = result.result["dryrun"]["resources"]

    def assert_handler_failed(resource, msg):
        changes = resources[resource]
        assert "changes" in changes
        changes = changes["changes"]
        assert "handler" in changes
        change = changes["handler"]
        assert change["current"] == "FAILED"
        assert change["desired"] == msg

    assert_handler_failed('test::Noprov[agent1,key=key1],v=%d' % version, "Unable to find a handler")
    assert_handler_failed('test::FailFast[agent1,key=key2],v=%d' % version, "Handler failed")
    assert_handler_failed('test::DoesNotExist[agent1,key=key2],v=%d' % version, "Resource Deserialization Failed")

    agent.stop()


@pytest.mark.gen_test
def test_send_events(resource_container, io_loop, environment, server, client):
    """
        Send and receive events within one agent
    """
    agentmanager = server.get_endpoint(SLICE_AGENT_MANAGER)

    resource_container.Provider.reset()
    agent = Agent(io_loop, hostname="node1", environment=environment, agent_map={"agent1": "localhost"},
                  code_loader=False)
    agent.add_end_point_name("agent1")
    agent.start()
    yield retry_limited(lambda: len(agentmanager.sessions) == 1, 10)

    version = int(time.time())

    res_id_1 = 'test::Resource[agent1,key=key1],v=%d' % version
    resources = [{'key': 'key1',
                  'value': 'value1',
                  'id': res_id_1,
                  'send_event': False,
                  'purged': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  'requires': ['test::Resource[agent1,key=key2],v=%d' % version],
                  },
                 {'key': 'key2',
                  'value': 'value2',
                  'id': 'test::Resource[agent1,key=key2],v=%d' % version,
                  'send_event': True,
                  'requires': [],
                  'purged': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  }
                 ]

    result = yield client.put_version(tid=environment, version=version, resources=resources, unknowns=[], version_info={})
    assert result.code == 200

    # do a deploy
    result = yield client.release_version(environment, version, True)
    assert result.code == 200

    result = yield client.get_version(environment, version)
    assert result.code == 200

    while (result.result["model"]["total"] - result.result["model"]["done"]) > 0:
        result = yield client.get_version(environment, version)
        yield gen.sleep(0.1)

    events = resource_container.Provider.getevents("agent1", "key1")
    assert len(events) == 1
    for res_id, res in events[0].items():
        assert res_id.agent_name == "agent1"
        assert res_id.attribute_value == "key2"
        assert res["status"] == const.ResourceState.deployed
        assert res["change"] == const.Change.created

    agent.stop()


@pytest.mark.gen_test
def test_send_events_cross_agent(resource_container, io_loop, environment, server, client):
    """
        Send and receive events over agents
    """
    agentmanager = server.get_endpoint(SLICE_AGENT_MANAGER)

    resource_container.Provider.reset()
    agent = Agent(io_loop, hostname="node1", environment=environment, agent_map={"agent1": "localhost"},
                  code_loader=False)
    agent.add_end_point_name("agent1")
    agent.start()
    yield retry_limited(lambda: len(agentmanager.sessions) == 1, 10)

    agent2 = Agent(io_loop, hostname="node2", environment=environment, agent_map={"agent2": "localhost"},
                   code_loader=False)
    agent2.add_end_point_name("agent2")
    agent2.start()
    yield retry_limited(lambda: len(agentmanager.sessions) == 2, 10)

    version = int(time.time())

    res_id_1 = 'test::Resource[agent1,key=key1],v=%d' % version
    resources = [{'key': 'key1',
                  'value': 'value1',
                  'id': res_id_1,
                  'send_event': False,
                  'purged': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  'requires': ['test::Resource[agent2,key=key2],v=%d' % version],
                  },
                 {'key': 'key2',
                  'value': 'value2',
                  'id': 'test::Resource[agent2,key=key2],v=%d' % version,
                  'send_event': True,
                  'requires': [],
                  'purged': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  }
                 ]

    result = yield client.put_version(tid=environment, version=version, resources=resources, unknowns=[], version_info={})
    assert result.code == 200

    # do a deploy
    result = yield client.release_version(environment, version, True)
    assert result.code == 200

    result = yield client.get_version(environment, version)
    assert result.code == 200

    while (result.result["model"]["total"] - result.result["model"]["done"]) > 0:
        result = yield client.get_version(environment, version)
        yield gen.sleep(0.1)

    assert resource_container.Provider.get("agent1", "key1") == "value1"
    assert resource_container.Provider.get("agent2", "key2") == "value2"

    events = resource_container.Provider.getevents("agent1", "key1")
    assert len(events) == 1
    for res_id, res in events[0].items():
        assert res_id.agent_name == "agent2"
        assert res_id.attribute_value == "key2"
        assert res["status"] == const.ResourceState.deployed
        assert res["change"] == const.Change.created

    agent.stop()
    agent2.stop()


@pytest.mark.gen_test(timeout=15)
def test_send_events_cross_agent_restart(resource_container, io_loop, environment, server, client):
    """
        Send and receive events over agents with agents starting after deploy
    """
    agentmanager = server.get_endpoint(SLICE_AGENT_MANAGER)

    resource_container.Provider.reset()
    agent2 = Agent(io_loop, hostname="node2", environment=environment, agent_map={"agent2": "localhost"},
                   code_loader=False)
    agent2.add_end_point_name("agent2")
    agent2.start()
    yield retry_limited(lambda: len(agentmanager.sessions) == 1, 10)

    version = int(time.time())

    res_id_1 = 'test::Resource[agent1,key=key1],v=%d' % version
    resources = [{'key': 'key1',
                  'value': 'value1',
                  'id': res_id_1,
                  'send_event': False,
                  'purged': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  'requires': ['test::Resource[agent2,key=key2],v=%d' % version],
                  },
                 {'key': 'key2',
                  'value': 'value2',
                  'id': 'test::Resource[agent2,key=key2],v=%d' % version,
                  'send_event': True,
                  'requires': [],
                  'purged': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  }
                 ]

    result = yield client.put_version(tid=environment, version=version, resources=resources, unknowns=[], version_info={})
    assert result.code == 200

    # do a deploy
    result = yield client.release_version(environment, version, True)
    assert result.code == 200

    result = yield client.get_version(environment, version)
    assert result.code == 200

    # wait for agent 2 to finish
    while (result.result["model"]["total"] - result.result["model"]["done"]) > 1:
        result = yield client.get_version(environment, version)
        yield gen.sleep(1)

    assert resource_container.Provider.get("agent2", "key2") == "value2"

    # start agent 1 and wait for it to finish
    Config.set("config", "agent-splay", "0")
    agent = Agent(io_loop, hostname="node1", environment=environment, agent_map={"agent1": "localhost"},
                  code_loader=False)
    agent.add_end_point_name("agent1")
    agent.start()
    yield retry_limited(lambda: len(agentmanager.sessions) == 2, 10)

    while (result.result["model"]["total"] - result.result["model"]["done"]) > 0:
        result = yield client.get_version(environment, version)
        yield gen.sleep(1)

    assert resource_container.Provider.get("agent1", "key1") == "value1"

    events = resource_container.Provider.getevents("agent1", "key1")
    assert len(events) == 1
    for res_id, res in events[0].items():
        assert res_id.agent_name == "agent2"
        assert res_id.attribute_value == "key2"
        assert res["status"] == const.ResourceState.deployed
        assert res["change"] == const.Change.created

    agent.stop()
    agent2.stop()


@pytest.mark.gen_test
def test_auto_deploy(io_loop, server, client, resource_container, environment):
    """
        dryrun and deploy a configuration model automatically
    """
    agentmanager = server.get_endpoint(SLICE_AGENT_MANAGER)

    resource_container.Provider.reset()
    agent = Agent(io_loop, hostname="node1", environment=environment, agent_map={"agent1": "localhost"},
                  code_loader=False)
    agent.add_end_point_name("agent1")
    agent.start()
    yield retry_limited(lambda: len(agentmanager.sessions) == 1, 10)

    resource_container.Provider.set("agent1", "key2", "incorrect_value")
    resource_container.Provider.set("agent1", "key3", "value")

    version = int(time.time())

    resources = [{'key': 'key1',
                  'value': 'value1',
                  'id': 'test::Resource[agent1,key=key1],v=%d' % version,
                  'send_event': False,
                  'purged': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  'requires': ['test::Resource[agent1,key=key2],v=%d' % version],
                  },
                 {'key': 'key2',
                  'value': 'value2',
                  'id': 'test::Resource[agent1,key=key2],v=%d' % version,
                  'send_event': False,
                  'requires': [],
                  'purged': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  },
                 {'key': 'key3',
                  'value': None,
                  'id': 'test::Resource[agent1,key=key3],v=%d' % version,
                  'send_event': False,
                  'requires': [],
                  'purged': True,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  }
                 ]

    # set auto deploy and push
    result = yield client.set_setting(environment, data.AUTO_DEPLOY, True)
    assert result.code == 200
    result = yield client.set_setting(environment, data.PUSH_ON_AUTO_DEPLOY, True)
    assert result.code == 200

    result = yield client.put_version(tid=environment, version=version, resources=resources, unknowns=[], version_info={})
    assert result.code == 200

    # check deploy
    result = yield client.get_version(environment, version)
    assert result.code == 200
    assert result.result["model"]["released"]
    assert result.result["model"]["total"] == 3
    assert result.result["model"]["result"] == "deploying"

    while (result.result["model"]["total"] - result.result["model"]["done"]) > 0:
        result = yield client.get_version(environment, version)
        yield gen.sleep(0.1)

    assert result.result["model"]["done"] == len(resources)

    assert resource_container.Provider.isset("agent1", "key1")
    assert resource_container.Provider.get("agent1", "key1") == "value1"
    assert resource_container.Provider.get("agent1", "key2") == "value2"
    assert not resource_container.Provider.isset("agent1", "key3")

    agent.stop()


@pytest.mark.gen_test(timeout=15)
def test_auto_deploy_no_splay(io_loop, server, client, resource_container, environment):
    """
        dryrun and deploy a configuration model automatically with agent autostart
    """
    resource_container.Provider.reset()
    env = yield data.Environment.get_by_id(uuid.UUID(environment))
    yield env.set(data.AUTOSTART_AGENT_MAP, {"agent1": ""})
    yield env.set(data.AUTOSTART_ON_START, True)

    version = int(time.time())

    resources = [{'key': 'key1',
                  'value': 'value1',
                  'id': 'test::Resource[agent1,key=key1],v=%d' % version,
                  'send_event': False,
                  'purged': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  'requires': ['test::Resource[agent1,key=key2],v=%d' % version],
                  },
                 ]

    # set auto deploy and push
    result = yield client.set_setting(environment, data.AUTO_DEPLOY, True)
    assert result.code == 200
    result = yield client.set_setting(environment, data.PUSH_ON_AUTO_DEPLOY, True)
    assert result.code == 200
    result = yield client.set_setting(environment, data.AUTOSTART_SPLAY, 0)
    assert result.code == 200

    result = yield client.put_version(tid=environment, version=version, resources=resources, unknowns=[], version_info={})
    assert result.code == 200

    # check deploy
    result = yield client.get_version(environment, version)
    assert result.code == 200
    assert result.result["model"]["released"]
    assert result.result["model"]["total"] == 1
    assert result.result["model"]["result"] == "deploying"

    # check if agent 1 is started by the server
    # deploy will fail because handler code is not uploaded to the server
    result = yield client.list_agents(tid=environment)
    assert result.code == 200

    while len(result.result["agents"]) == 0 or result.result["agents"][0]["state"] == "down":
        result = yield client.list_agents(tid=environment)
        yield gen.sleep(0.1)

    assert len(result.result["agents"]) == 1
    assert result.result["agents"][0]["name"] == "agent1"


@pytest.mark.gen_test(timeout=15)
def test_autostart_mapping(io_loop, server, client, resource_container, environment):
    """
        Test autostart mapping and restart agents when the map is modified
    """
    resource_container.Provider.reset()
    env = yield data.Environment.get_by_id(uuid.UUID(environment))
    yield env.set(data.AUTOSTART_AGENT_MAP, {"agent1": ""})
    yield env.set(data.AUTO_DEPLOY, True)
    yield env.set(data.PUSH_ON_AUTO_DEPLOY, True)
    yield env.set(data.AUTOSTART_SPLAY, 0)
    yield env.set(data.AUTOSTART_ON_START, True)

    version = int(time.time())

    resources = [{'key': 'key1',
                  'value': 'value1',
                  'id': 'test::Resource[agent1,key=key1],v=%d' % version,
                  'send_event': False,
                  'purged': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  'requires': [],
                  },
                 {'key': 'key1',
                  'value': 'value1',
                  'id': 'test::Resource[agent2,key=key1],v=%d' % version,
                  'send_event': False,
                  'purged': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  'requires': [],
                  },
                 ]

    result = yield client.put_version(tid=environment, version=version, resources=resources, unknowns=[], version_info={})
    assert result.code == 200

    # check deploy
    result = yield client.get_version(environment, version)
    assert result.code == 200
    assert result.result["model"]["released"]
    assert result.result["model"]["total"] == 2
    assert result.result["model"]["result"] == "deploying"

    result = yield client.list_agents(tid=environment)
    assert result.code == 200

    while len([x for x in result.result["agents"] if x["state"] == "up"]) < 1:
        result = yield client.list_agents(tid=environment)
        yield gen.sleep(0.1)

    assert len(result.result["agents"]) == 2
    assert len([x for x in result.result["agents"] if x["state"] == "up"]) == 1

    result = yield client.set_setting(environment, data.AUTOSTART_AGENT_MAP, {"agent1": "", "agent2": ""})
    assert result.code == 200

    result = yield client.list_agents(tid=environment)
    assert result.code == 200
    while len([x for x in result.result["agents"] if x["state"] == "up"]) < 2:
        result = yield client.list_agents(tid=environment)
        yield gen.sleep(0.1)


@pytest.mark.gen_test(timeout=15)
def test_autostart_clear_environment(io_loop, server_multi, client_multi, resource_container, environment):
    """
        Test clearing an environment with autostarted agents. After clearing, autostart should still work
    """
    resource_container.Provider.reset()
    env = yield data.Environment.get_by_id(uuid.UUID(environment))
    yield env.set(data.AUTOSTART_AGENT_MAP, {"agent1": ""})
    yield env.set(data.AUTO_DEPLOY, True)
    yield env.set(data.PUSH_ON_AUTO_DEPLOY, True)
    yield env.set(data.AUTOSTART_SPLAY, 0)
    yield env.set(data.AUTOSTART_ON_START, True)

    version = int(time.time())

    resources = [{'key': 'key1',
                  'value': 'value1',
                  'id': 'test::Resource[agent1,key=key1],v=%d' % version,
                  'send_event': False,
                  'purged': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  'requires': [],
                  }
                 ]

    client = client_multi
    result = yield client.put_version(tid=environment, version=version, resources=resources, unknowns=[], version_info={})
    assert result.code == 200

    # check deploy
    result = yield client.get_version(environment, version)
    assert result.code == 200
    assert result.result["model"]["released"]
    assert result.result["model"]["total"] == 1
    assert result.result["model"]["result"] == "deploying"

    result = yield client.list_agents(tid=environment)
    assert result.code == 200

    while len([x for x in result.result["agents"] if x["state"] == "up"]) < 1:
        result = yield client.list_agents(tid=environment)
        yield gen.sleep(0.1)

    assert len(result.result["agents"]) == 1
    assert len([x for x in result.result["agents"] if x["state"] == "up"]) == 1

    # clear environment
    yield client.clear_environment(environment)

    items = yield data.ConfigurationModel.get_list()
    assert len(items) == 0
    items = yield data.Resource.get_list()
    assert len(items) == 0
    items = yield data.ResourceAction.get_list()
    assert len(items) == 0
    items = yield data.Code.get_list()
    assert len(items) == 0
    items = yield data.Agent.get_list()
    assert len(items) == 0
    items = yield data.AgentInstance.get_list()
    assert len(items) == 0
    items = yield data.AgentProcess.get_list()
    assert len(items) == 0

    # Do a deploy again
    version = int(time.time())

    resources = [{'key': 'key1',
                  'value': 'value1',
                  'id': 'test::Resource[agent1,key=key1],v=%d' % version,
                  'send_event': False,
                  'purged': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  'requires': [],
                  }
                 ]

    result = yield client.put_version(tid=environment, version=version, resources=resources, unknowns=[], version_info={})
    assert result.code == 200

    # check deploy
    result = yield client.get_version(environment, version)
    assert result.code == 200
    assert result.result["model"]["released"]
    assert result.result["model"]["total"] == 1
    assert result.result["model"]["result"] == "deploying"

    result = yield client.list_agents(tid=environment)
    assert result.code == 200

    while len([x for x in result.result["agents"] if x["state"] == "up"]) < 1:
        result = yield client.list_agents(tid=environment)
        yield gen.sleep(0.1)

    assert len(result.result["agents"]) == 1
    assert len([x for x in result.result["agents"] if x["state"] == "up"]) == 1


@pytest.mark.gen_test
def test_export_duplicate(resource_container, snippetcompiler):
    """
        The exported should provide a compilation error when a resource is defined twice in a model
    """
    snippetcompiler.setup_for_snippet("""
        import test

        test::Resource(key="test", value="foo")
        test::Resource(key="test", value="bar")
    """)

    with pytest.raises(CompilerException) as exc:
        snippetcompiler.do_export()

    assert "exists more than once in the configuration model" in str(exc.value)


@pytest.mark.gen_test(timeout=90)
def test_server_recompile(server_multi, client_multi, environment_multi):
    """
        Test a recompile on the server and verify recompile triggers
    """
    config.Config.set("server", "auto-recompile-wait", "0")
    client = client_multi
    server = server_multi
    environment = environment_multi

    @gen.coroutine
    def wait_for_version(cnt):
        # Wait until the server is no longer compiling
        # wait for it to finish
        code = 200
        while code == 200:
            compiling = yield client.is_compiling(environment)
            code = compiling.code
            yield gen.sleep(1)

        # wait for it to appear
        versions = yield client.list_versions(environment)

        while versions.result["count"] < cnt:
            versions = yield client.list_versions(environment)

        return versions.result

    project_dir = os.path.join(server.get_endpoint("server")._server_storage["environments"], str(environment))
    project_source = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data", "project")

    shutil.copytree(project_source, project_dir)
    subprocess.check_output(["git", "init"], cwd=project_dir)
    subprocess.check_output(["git", "add", "*"], cwd=project_dir)
    subprocess.check_output(["git", "config", "user.name", "Unit"], cwd=project_dir)
    subprocess.check_output(["git", "config", "user.email", "unit@test.example"], cwd=project_dir)
    subprocess.check_output(["git", "commit", "-m", "unit test"], cwd=project_dir)

    # add main.cf
    with open(os.path.join(project_dir, "main.cf"), "w") as fd:
        fd.write("""
        host = std::Host(name="test", os=std::linux)
        std::ConfigFile(host=host, path="/etc/motd", content="1234")
""")

    # request a compile
    yield client.notify_change(environment)

    versions = yield wait_for_version(1)
    assert versions["versions"][0]["total"] == 1
    assert versions["versions"][0]["version_info"]["export_metadata"]["type"] == "api"

    # get compile reports
    reports = yield client.get_reports(environment)
    assert len(reports.result["reports"]) == 1

    # set a parameter without requesting a recompile
    yield client.set_param(environment, id="param1", value="test", source="plugin")
    versions = yield wait_for_version(1)
    assert versions["count"] == 1

    # set a new parameter and request a recompile
    yield client.set_param(environment, id="param2", value="test", source="plugin", recompile=True)
    versions = yield wait_for_version(2)
    assert versions["versions"][0]["version_info"]["export_metadata"]["type"] == "param"
    assert versions["count"] == 2

    # update the parameter to the same value -> no compile
    yield client.set_param(environment, id="param2", value="test", source="plugin", recompile=True)
    versions = yield wait_for_version(2)
    assert versions["count"] == 2

    # update the parameter to a new value
    yield client.set_param(environment, id="param2", value="test2", source="plugin", recompile=True)
    versions = yield wait_for_version(3)
    assert versions["count"] == 3


class ResourceProvider(object):

    def __init__(self, index, name, producer, state=None):
        self.name = name
        self.producer = producer
        self.state = state
        self.index = index

    def get_resource(self,
                     resource_container: ResourceContainer,
                     agent: str,
                     key: str,
                     version: str,
                     requires: List[str]) -> Tuple[Dict[str, str], Optional[ResourceState]]:
        base = {'key': key,
                'value': 'value1',
                'id': 'test::Resource[%s,key=%s],v=%d' % (agent, key, version),
                'send_event': True,
                'purged': False,
                'state_id': '',
                'allow_restore': True,
                'allow_snapshot': True,
                'requires': requires,
                }

        self.producer(resource_container.Provider, agent, key)

        state = None
        if self.state is not None:
            state = ('test::Resource[%s,key=%s]' % (agent, key), self.state)

        return base, state

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name


# for events, self is the consuming node
# dep is the producer/required node
self_states = [
    ResourceProvider(0, "skip", lambda p, a, k:p.set_skip(a, k, 1)),
    ResourceProvider(1, "fail", lambda p, a, k:p.set_fail(a, k, 1)),
    ResourceProvider(2, "success", lambda p, a, k: None),
    ResourceProvider(3, "undefined", lambda p, a, k: None, const.ResourceState.undefined),
]

dep_states = [
    ResourceProvider(0, "skip", lambda p, a, k:p.set_skip(a, k, 1)),
    ResourceProvider(1, "fail", lambda p, a, k:p.set_fail(a, k, 1)),
    ResourceProvider(2, "success", lambda p, a, k: None),
]


def make_matrix(matrix, valueparser):
    """
    Expect matrix of the form

        header1    header2     header3
    row1    y    y    n
    """
    unparsed = [
        [v for v in row.split()][1:]
        for row in matrix.strip().split("\n")
    ][1:]

    return [[valueparser(nsv) for nsv in nv] for nv in unparsed]


# self state on X axis
# dep state on the Y axis
dorun = make_matrix("""
        skip    fail    success    undef
skip    n    n    n    n
fail    n    n    n    n
succ    y    y    y    n
""", lambda x: x == "y")

dochange = make_matrix("""
        skip    fail    success    undef
skip    n    n    n    n
fail    n    n    n    n
succ    n    n    y    n
""", lambda x: x == "y")

doevents = make_matrix("""
        skip    fail    success    undef
skip    2    2    2    0
fail    2    2    2    0
succ    2    2    2    0
""", lambda x: int(x))


@pytest.mark.parametrize("self_state", self_states, ids=lambda x: x.name)
@pytest.mark.parametrize("dep_state", dep_states, ids=lambda x: x.name)
@pytest.mark.gen_test
def test_deploy_and_events(io_loop, client, server, environment, resource_container, self_state, dep_state):

    agentmanager = server.get_endpoint(SLICE_AGENT_MANAGER)

    resource_container.Provider.reset()
    agent = Agent(io_loop, hostname="node1", environment=environment, agent_map={"agent1": "localhost"},
                  code_loader=False)
    agent.add_end_point_name("agent1")
    agent.start()
    yield retry_limited(lambda: len(agentmanager.sessions) == 1, 10)

    version = int(time.time())

    (dep, dep_status) = dep_state.get_resource(resource_container, "agent1", "key2", version, [])
    (own, own_status) = self_state.get_resource(resource_container, "agent1", "key3", version, [
        'test::Resource[agent1,key=key2],v=%d' % version, 'test::Resource[agent1,key=key1],v=%d' % version])

    resources = [{'key': 'key1',
                  'value': 'value1',
                  'id': 'test::Resource[agent1,key=key1],v=%d' % version,
                  'send_event': True,
                  'purged': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  'requires': [],
                  },
                 dep,
                 own
                 ]

    status = {x[0]: x[1] for x in [dep_status, own_status] if x is not None}
    result = yield client.put_version(tid=environment, version=version, resources=resources, resource_state=status,
                                      unknowns=[], version_info={})
    assert result.code == 200

    # do a deploy
    result = yield client.release_version(environment, version, True)
    assert result.code == 200
    assert not result.result["model"]["deployed"]
    assert result.result["model"]["released"]
    assert result.result["model"]["total"] == 3
    assert result.result["model"]["result"] == "deploying"

    result = yield client.get_version(environment, version)
    assert result.code == 200

    while (result.result["model"]["total"] - result.result["model"]["done"]) > 0:
        result = yield client.get_version(environment, version)
        yield gen.sleep(0.1)

    assert result.result["model"]["done"] == len(resources)

    # verify against result matrices
    assert dorun[dep_state.index][self_state.index] == (resource_container.Provider.readcount("agent1", "key3") > 0)
    assert dochange[dep_state.index][self_state.index] == (resource_container.Provider.changecount("agent1", "key3") > 0)

    events = resource_container.Provider.getevents("agent1", "key3")
    expected_events = doevents[dep_state.index][self_state.index]
    if expected_events == 0:
        assert len(events) == 0
    else:
        assert len(events) == 1
        assert len(events[0]) == expected_events


@pytest.mark.gen_test
def test_deploy_and_events_failed(io_loop, client, server, environment, resource_container):
    agentmanager = server.get_endpoint(SLICE_AGENT_MANAGER)

    resource_container.Provider.reset()
    agent = Agent(io_loop, hostname="node1", environment=environment, agent_map={"agent1": "localhost"},
                  code_loader=False)
    agent.add_end_point_name("agent1")
    agent.start()
    yield retry_limited(lambda: len(agentmanager.sessions) == 1, 10)

    version = int(time.time())

    resources = [{'key': 'key1',
                  'value': 'value1',
                  'id': 'test::Resource[agent1,key=key1],v=%d' % version,
                  'send_event': True,
                  'purged': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  'requires': [],
                  },
                 {'key': 'key2',
                  'value': 'value1',
                  'id': 'test::BadEvents[agent1,key=key2],v=%d' % version,
                  'send_event': True,
                  'purged': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  'requires': ['test::Resource[agent1,key=key1],v=%d' % version],
                  },
                 ]

    result = yield client.put_version(tid=environment, version=version, resources=resources, resource_state={},
                                      unknowns=[], version_info={})
    assert result.code == 200

    # do a deploy
    result = yield client.release_version(environment, version, True)
    assert result.code == 200
    assert not result.result["model"]["deployed"]
    assert result.result["model"]["released"]
    assert result.result["model"]["total"] == 2
    assert result.result["model"]["result"] == "deploying"

    result = yield client.get_version(environment, version)
    assert result.code == 200

    while (result.result["model"]["total"] - result.result["model"]["done"]) > 0:
        result = yield client.get_version(environment, version)
        yield gen.sleep(0.1)

    assert result.result["model"]["done"] == len(resources)


dep_states_reload = [
    ResourceProvider(0, "skip", lambda p, a, k:p.set_skip(a, k, 1)),
    ResourceProvider(0, "fail", lambda p, a, k:p.set_fail(a, k, 1)),
    ResourceProvider(0, "nochange", lambda p, a, k: p.set(a, k, "value1")),
    ResourceProvider(1, "changed", lambda p, a, k: None)
]


@pytest.mark.parametrize("dep_state", dep_states_reload, ids=lambda x: x.name)
@pytest.mark.gen_test(timeout=5000)
def test_reload(io_loop, client, server, environment, resource_container, dep_state):

    agentmanager = server.get_endpoint(SLICE_AGENT_MANAGER)

    resource_container.Provider.reset()
    agent = Agent(io_loop, hostname="node1", environment=environment, agent_map={"agent1": "localhost"},
                  code_loader=False)
    agent.add_end_point_name("agent1")
    agent.start()
    yield retry_limited(lambda: len(agentmanager.sessions) == 1, 10)

    version = int(time.time())

    (dep, dep_status) = dep_state.get_resource(resource_container, "agent1", "key1", version, [])

    resources = [{'key': 'key2',
                  'value': 'value1',
                  'id': 'test::Resource[agent1,key=key2],v=%d' % version,
                  'send_event': True,
                  'purged': False,
                  'state_id': '',
                  'allow_restore': True,
                  'allow_snapshot': True,
                  'requires': ['test::Resource[agent1,key=key1],v=%d' % version],
                  },
                 dep
                 ]

    status = {x[0]: x[1] for x in [dep_status] if x is not None}
    result = yield client.put_version(tid=environment, version=version, resources=resources, resource_state=status,
                                      unknowns=[], version_info={})
    assert result.code == 200

    # do a deploy
    result = yield client.release_version(environment, version, True)
    assert result.code == 200
    assert not result.result["model"]["deployed"]
    assert result.result["model"]["released"]
    assert result.result["model"]["total"] == 2
    assert result.result["model"]["result"] == "deploying"

    result = yield client.get_version(environment, version)
    assert result.code == 200

    while (result.result["model"]["total"] - result.result["model"]["done"]) > 0:
        result = yield client.get_version(environment, version)
        yield gen.sleep(0.1)

    assert result.result["model"]["done"] == len(resources)

    assert dep_state.index == resource_container.Provider.reloadcount("agent1", "key2")
